import builtins
import os
import re
import string
import sys
import unicodedata
from heapq import nlargest

import fastnumbers
import pyspark
from multipledispatch import dispatch
from pyspark.ml.feature import Imputer, QuantileDiscretizer
from pyspark.ml.feature import VectorAssembler
from pyspark.ml.linalg import Vectors, VectorUDT
from pyspark.ml.stat import Correlation
from pyspark.sql import DataFrame
from pyspark.sql import functions as F
from pyspark.sql.functions import when

import optimus.helpers.functions_spark
from optimus import ROOT_DIR
from optimus.engines.base.distributed.columns import DistributedBaseColumns
from optimus.engines.base.meta import Meta
# Helpers
from optimus.engines.base.pandas.columns import PandasBaseColumns
from optimus.engines.spark.ml.encoding import index_to_string as ml_index_to_string
from optimus.engines.spark.ml.encoding import string_to_index as ml_string_to_index
from optimus.helpers.check import has_, is_column_a, is_spark_dataframe
from optimus.helpers.columns import get_output_cols, parse_columns, check_column_numbers, validate_columns_names, \
    name_col, prepare_columns
from optimus.helpers.constants import RELATIVE_ERROR, Actions
from optimus.helpers.converter import format_dict
from optimus.helpers.core import val_to_list, one_list_to_val
from optimus.helpers.functions \
    import filter_list, create_buckets
from optimus.helpers.functions_spark import append as append_df
from optimus.helpers.logger import logger
from optimus.helpers.parser import parse_python_dtypes, compress_list
from optimus.helpers.raiseit import RaiseIt
from optimus.helpers.types import *
from optimus.profiler.functions import fill_missing_var_types

# Add the directory containing your module to the Python path (wants absolute paths)
sys.path.append(os.path.abspath(ROOT_DIR))

# To use this functions as a Spark udf function we need to load it using addPyFile because the file can not be loaded
# as python module because it generate a pickle error.
# from infer import Infer

from optimus.infer import is_, is_type, is_function, is_list_value, is_tuple, is_list_of_str, \
    is_list_of_tuples, is_one_element, is_num_or_str, is_numeric, is_str, is_int
# from optimus.infer_spark import SPARK_DTYPES_TO_INFERRED, parse_spark_class_dtypes, is_list_of_spark_dataframes
# NUMERIC_TYPES, NOT_ARRAY_TYPES, STRING_TYPES, ARRAY_TYPES
from optimus.engines.spark.audf import filter_row_by_data_type as fbdt

# Functions

ENGINE = "spark"


class Cols(PandasBaseColumns, DistributedBaseColumns):

    def _names(self):
        return list(self.root.data.columns)

    @dispatch(str, object)
    def append(self, col_name=None, value=None):
        """
        Append a column to a Dataframe
        :param col_name: Name of the new column
        :param value: List of data values
        :return:
        """

        def lit_array(_value):
            temp = []
            for v in _value:
                temp.append(F.lit(v))
            return F.array(temp)

        dfd = self.root.data

        if is_num_or_str(value):
            value = F.lit(value)
        elif is_list_value(value):
            value = lit_array(value)
        elif is_tuple(value):
            value = lit_array(list(value))

        if is_(value, F.Column):
            dfd = dfd.withColumn(col_name, value)

        return self.root.new(dfd)

    @dispatch((list, pyspark.sql.dataframe.DataFrame))
    def append(self, cols_values=None):
        """
        Append a column or a Dataframe to a Dataframe
        :param cols_values: New Column Names and data values
        :type cols_values: List of tuples
        :return:
        """
        df = self.root
        df_result = None
        if is_list_of_tuples(cols_values):

            for c in cols_values:
                col_name = c[0]
                value = c[1]
                df_result = optimus.helpers.functions_spark.append(col_name, value)

        elif is_list_of_spark_dataframes(cols_values) or is_spark_dataframe(cols_values):
            cols_values = val_to_list(cols_values)
            cols_values.insert(0, df)
            df_result = append_df(cols_values, like="columns")

        else:
            RaiseIt.type_error(cols_values, ["list of tuples", "dataframes"])

        return df_result

    # @staticmethod
    # def select(columns="*", regex=None, data_type=None, invert=False):
    #     """
    #     Select columns using index, column name, regex to data type
    #     :param columns:
    #     :param regex: Regular expression to filter the columns
    #     :param data_type: Data type to be filtered for
    #     :param invert: Invert the selection
    #     :return:
    #     """
    #     df = self
    #     columns = parse_columns(df, columns, is_regex=regex, filter_by_column_types=data_type, invert=invert)
    #     if columns is not None:
    #         df = df.select(columns)
    #         # Metadata get lost when using select(). So we copy here again.
    #         df.meta = Meta.action(df.meta, None)
    #
    #     else:
    #         df = None
    #
    #     return df

    def copy(self, input_cols, output_cols=None, columns=None):
        """
        Copy one or multiple columns
        :param input_cols: Source column to be copied
        :param output_cols: Destination column
        :param columns: tuple of column [('column1','column_copy')('column1','column1_copy')()]
        :return:
        """
        df = self.root

        if is_list_of_str(columns):
            output_cols = [col_name + "_copy" for col_name in columns]
            output_cols = get_output_cols(columns, output_cols)
            columns = zip(columns, output_cols)
        else:
            input_cols = parse_columns(dfd, input_cols)
            output_cols = get_output_cols(input_cols, output_cols)

            columns = list(zip(input_cols, output_cols))

        for input_col, output_col in columns:
            current_meta = dfd.meta.get()
            dfd = df.data.withColumn(output_col, F.col(input_col))
            meta.set(value=current_meta)
            meta = Meta.action(meta, Actions.COPY.value, (input_col, output_col))

        return self.root.new(dfd, meta=meta)

    @staticmethod
    def to_timestamp(input_cols, date_format=None, output_cols=None):
        """
        Convert a string to timestamp
        :param input_cols:
        :param output_cols:
        :param date_format:
        :return:
        """
        df = self.root

        input_cols = parse_columns(df, input_cols)
        output_cols = get_output_cols(input_cols, output_cols)

        for input_col, output_col in zip(input_cols, output_cols):
            df = df.withColumn(output_col, F.to_timestamp(input_col, date_format))
        return df

    # def apply(self, input_cols, func=None, func_return_type=None, args=None, func_type=None, when=None,
    #           filter_col_by_dtypes=None, output_cols=None, skip_output_cols_processing=False,
    #           meta_action="apply", default=None, mode=None):
    #     """
    #     Apply a function using pandas udf or udf if apache arrow is not available
    #     :param input_cols: Columns in which the function is going to be applied
    #     :param output_cols: Columns in which the transformed data will saved
    #     :param func: Functions to be applied to a columns. The declaration must have always 2 params.
    #         def func(value, args):
    #     :param func_return_type: function return type. This is required by UDF and Pandas UDF.
    #     :param args: Arguments to be passed to the function
    #     :param func_type: pandas_udf or udf. If none try to use pandas udf (Pyarrow needed)
    #     :param when: A expression to better control when the function is going to be applied
    #     :param filter_col_by_dtypes: Only apply the filter to specific type of value ,integer, float, string or bool
    #     :param skip_output_cols_processing: In some special cases we do not want apply() to construct the output columns.
    #     True or False
    #     :param meta_action: Metadata transformation to a dataframe
    #     """
    #     func_return_type = {str: "string", int: "int", float: "float", bool: "boolean"}.get(func_return_type)
    #     df = self.root
    #     columns = prepare_columns(df, input_cols, output_cols, filter_by_column_types=filter_col_by_dtypes,
    #                               accepts_missing_cols=True, default=default)
    #
    #     def expr(_when):
    #         main_query = audf(input_col, func, func_return_type, args, func_type)
    #         if when is not None:
    #             # Use the data type to filter the query
    #             main_query = F.when(_when, main_query).otherwise(F.col(input_col))
    #
    #         return main_query
    #
    #     dfd = df.data
    #     meta = df.meta
    #     for input_col, output_col in columns:
    #         # print("expr(when)",type(expr(when)),expr(when))
    #         dfd = dfd.withColumn(output_col, expr(when))
    #         meta = Meta.action(meta, meta_action, output_col)
    #         # self.root.meta.preserve(self, meta_action, output_col)
    #     # dfd = self.root.new(dfd)
    #
    #     return self.root.new(dfd)

    @staticmethod
    def apply_by_data_type(columns, func, func_return_type, args=None, func_type=None, data_type=None):
        """
        Apply a function using pandas udf or udf if apache arrow is not available
        :param columns: Columns in which the function is going to be applied
        :param func: Functions to be applied to a columns
        :param func_return_type
        :param args:
        :param func_type: pandas_udf or udf. If none try to use pandas udf (Pyarrow needed)
        :param data_type:
        :return:
        """
        df = self.root
        columns = parse_columns(df, columns)
        for col_name in columns:
            df = df.cols.apply(col_name, func=func, func_return_type=func_return_type, args=args,
                               func_type=func_type,
                               when=fbdt(col_name, data_type))
        return df

    # TODO: Maybe we could merge this with apply()
    def set(self, output_cols=None, value=None, where=None, default=None, eval_value=False):
        """
        Set a column value using a number, string or a expression.
        :param where:
        :param value:
        :param output_cols:
        :param default:
        :return:
        """
        df = self.root

        columns = parse_columns(df, output_cols, accepts_missing_cols=True)
        check_column_numbers(columns, 1)

        if is_list_value(value):
            expr = F.array([F.lit(x) for x in value])
        elif is_numeric(value):
            expr = F.lit(value)
        elif is_str(value):
            expr = F.expr(value)
        else:
            RaiseIt.value_error(value, ["numeric", "list", "hive expression"])

        df = df.withColumn(output_cols, expr)
        df.meta = Meta.action(df.meta, None, Actions.SET.value, columns)
        return df

    def create_exprs(self, columns, funcs, *args, df=None):
        """
        Helper function to apply multiple columns expression to multiple columns
        :param columns:
        :param funcs:
        :param args:
        :return:
        """
        df = self.root
        columns = parse_columns(df, columns)
        funcs = val_to_list(funcs)
        exprs = []

        for col_name in columns:
            for func in funcs:
                exprs.append((func, (col_name, *args, df)))
        # print("exprs",exprs)
        # df = self

        # Std, kurtosis, mean, skewness and other agg functions can not process date columns.
        filters = {"date": [F.stddev, F.kurtosis, F.mean, F.skewness, F.sum, F.variance, F.approx_count_distinct,
                            self.F.zeros_agg],
                   "array": [F.stddev, F.kurtosis, F.mean, F.skewness, F.sum, F.variance, F.approx_count_distinct,
                             self.F.zeros_agg],
                   "timestamp": [F.stddev, F.kurtosis, F.mean, F.skewness, F.sum, F.variance,
                                 F.approx_count_distinct,
                                 self.F.zeros_agg, self.F.percentile],
                   "null": [F.stddev, F.kurtosis, F.mean, F.skewness, F.sum, F.variance, F.approx_count_distinct,
                            self.F.zeros_agg],
                   "boolean": [F.stddev, F.kurtosis, F.mean, F.skewness, F.sum, F.variance, F.approx_count_distinct,
                               self.F.zeros_agg],
                   "binary": [F.stddev, F.kurtosis, F.mean, F.skewness, F.sum, F.variance, F.approx_count_distinct,
                              self.F.zeros_agg]
                   }

        def _filter(_col_name, _func):
            for data_type, func_filter in filters.items():
                for f in func_filter:
                    if (_func == f) and (is_column_a(df, _col_name, data_type)):
                        return True
            return False

        beauty_col_names = {"hist_agg": "hist", "percentile": "percentile", "zeros_agg": "zeros",
                            "count_na_agg": "count_na", "range_agg": "range", "count_uniques_agg": "count_uniques"}

        def _beautify_col_names(_func):
            if _func.__name__ in beauty_col_names:
                func_name = beauty_col_names[_func.__name__]
            else:
                func_name = _func.__name__
            return func_name

        def _agg_exprs(_funcs):
            _exprs = []
            for f in _funcs:
                _func = f[0]
                _args = f[1]
                _col_name = _args[0]

                if not _filter(_col_name, _func):
                    agg = _func(*_args)
                    if agg is not None:
                        func_name = _beautify_col_names(_func)
                        _exprs.append(agg.alias(func_name + "_" + _col_name))

            return _exprs

        r = _agg_exprs(exprs)

        return r

    @staticmethod
    def range(columns):
        """
        Return the range form the min to the max value
        :param columns: '*', list of columns names or a single column name.
        :return:
        """

        return Cols.agg_exprs(columns, self.root.functions.range_agg)

    # Descriptive Analytics
    @staticmethod
    # TODO: implement double MAD http://eurekastatistics.com/using-the-median-absolute-deviation-to-find-outliers/
    def mad(columns, relative_error=RELATIVE_ERROR, more=None):
        """
        Return the Median Absolute Deviation
        :param columns: Column to be processed
        :param more: Return some extra computed values (Median).
        :param relative_error: Relative error calculating the media
        :return:
        """
        columns = parse_columns(
            self.root, columns, filter_by_column_types=self.root.constants.NUMERIC_TYPES)
        check_column_numbers(columns, "*")

        result = {}
        for col_name in columns:

            _mad = {}
            median_value = self.root.cols.median(col_name, relative_error)
            mad_value = self.root.data.withColumn(col_name, F.abs(F.col(col_name) - median_value)) \
                .cols.median(col_name, relative_error)

            if more:
                _mad = {"mad": mad_value, "median": median_value}
            else:
                _mad = {"mad": mad_value}

            result[col_name] = _mad

        return format_dict(result)

    @staticmethod
    def sum(columns):
        """
        Return the sum of a column dataframe
        :param columns: '*', list of columns names or a single column name.
        :return:
        """
        columns = parse_columns(
            self.root, columns, filter_by_column_types=self.root.constants.NUMERIC_TYPES)
        check_column_numbers(columns, "*")

        return format_dict(Cols.agg_exprs(columns, F.sum))

    @staticmethod
    def reverse(columns):
        pass

    @staticmethod
    def mode(columns):
        """
        Return the the column mode
        :param columns: '*', list of columns names or a single column name.
        :return:
        """

        columns = parse_columns(self.root, columns)
        mode_result = []

        for col_name in columns:
            count = self.groupBy(col_name).count()
            mode_df = count.join(
                count.agg(F.max("count").alias("max_")), F.col("count") == F.col("max_")
            )

            mode_df = mode_df.cache()
            # if none of the values are repeated we not have mode
            mode_list = (mode_df
                         .rows.select(mode_df["count"] > 1)
                         .cols.select(col_name)
                         .collect())

            mode_result.append({col_name: filter_list(mode_list)})

        return format_dict(mode_result)

    # String Operations

    def remove(self, columns, search=None, search_by="chars", output_cols=None):
        """
        Remove chars or words
        :param columns: '*', list of columns names or a single column name.
        :param search: values to look at to be replaced
        :param search_by: Match substring or words
        :param output_cols:
        :return:
        """
        return self.replace(columns, search, "", search_by, output_cols)

    def normalize_chars(self, input_cols="*", output_cols=None):
        """
        Remove accents in specific columns
        :param input_cols: '*', list of columns names or a single column name.
        :param output_cols:
        :return:
        """

        def _normalize_chars(value):
            value = str(value)

            # first, normalize strings:
            nfkd_str = unicodedata.normalize('NFKD', value)

            # Keep chars that has no other char combined (i.e. accents chars)
            with_out_accents = u"".join([c for c in nfkd_str if not unicodedata.combining(c)])
            return with_out_accents

        df = self.apply(input_cols, _normalize_chars, str, output_cols=output_cols,
                        meta_action=Actions.REMOVE_ACCENTS.value)
        return df

    def remove_special_chars(self, input_cols="*", output_cols=None):
        """
        Reference https://stackoverflow.com/questions/265960/best-way-to-strip-punctuation-from-a-string-in-python
        This method remove special characters (i.e. !”#$%&/()=?) in columns of dataFrames.
        :param input_cols: '*', list of columns names or a single column name.
        :param output_cols:
        :return:
        """

        input_cols = parse_columns(
            self.root, input_cols, filter_by_column_types=self.root.constants.STRING_TYPES)
        check_column_numbers(input_cols, "*")

        df = self.replace(input_cols, [f"\\{s}" for s in string.punctuation], "", "chars", output_cols=output_cols)
        return df

    def remove_white_spaces(self, input_cols="*", output_cols=None):
        """
        Remove all the white spaces from a string
        :param input_cols:
        :param output_cols:
        :return:
        """

        def _remove_white_spaces(col_name, args):
            return F.regexp_replace(F.col(col_name), " ", "")

        return self.apply(input_cols, _remove_white_spaces, output_cols=output_cols,
                          filter_col_by_dtypes=self.root.constants.NOT_ARRAY_TYPES,
                          meta_action=Actions.REMOVE_WHITE_SPACES.value)

    @staticmethod
    def format_date(input_cols, current_format=None, output_format=None, output_cols=None):
        """
        Transform a column date to a specified format
        :param input_cols: Columns to be transformed.
        :param output_cols: Output columns
        :param current_format: Current_format is the current string dat format of columns specified. Of course,
                                all columns specified must have the same format. Otherwise the function is going
                                to return tons of null values because the transformations in the columns with
                                different formats will fail. For example "yyyy/MM/dd"
        :param output_format: Output date string format to be expected. For example "dd-MM-YYYY"
        """

        def _date_format(col_name, attr):
            _current_format = attr[0]
            _output_format = attr[1]

            return F.date_format(F.unix_timestamp(col_name, _current_format).cast("timestamp"),
                                 _output_format).alias(
                col_name)

        # Asserting if column if in dataFrame:
        df = Cols.apply(input_cols, _date_format, args=[current_format, output_format], output_cols=output_cols)

        return df

    @staticmethod
    def years_between(input_cols, date_format=None, output_cols=None):
        """
        Years between a date and now
        :param input_cols: Input columns
        :param output_cols:
        :param date_format: String format date of the column provided. For example "yyyyMMdd"
        """

        # Output format date
        format_dt = "yyyy-MM-dd"

        def _years_between(col_name, attr):
            _date_format = attr[0]

            return F.format_number(
                F.abs(
                    F.months_between(
                        F.date_format(
                            F.unix_timestamp(
                                col_name,
                                _date_format).cast("timestamp"),
                            format_dt),
                        F.current_date()) / 12), 4) \
                .alias(
                col_name)

        # output_cols = get_output_cols(input_cols, output_cols)

        df = Cols.apply(input_cols, func=_years_between, args=[date_format],
                        filter_col_by_dtypes=self.root.constants.NOT_ARRAY_TYPES, output_cols=output_cols)

        df = df.cols.cast(output_cols, "float")

        return df

    def match(self, input_cols, regex=None, value=None, output_cols=None):
        """
        Use a Regex to replace values
        :param input_cols: '*', list of columns names or a single column name.
        :param output_cols:
        :param regex: values to look at to be replaced
        :param value: new value to replace the old one
        :return:
        """

        # if regex or normal replace we use regexp or replace functions
        def _match(_input_cols, attr):
            _search = attr[0]
            return F.when(F.col(_input_cols).rlike(_search), True).otherwise(False)

        return self.apply(input_cols, func=_match, args=(regex,), func_type="column_expr", func_return_type=str,
                          output_cols=output_cols,
                          meta_action=Actions.REPLACE_REGEX.value)

    def replace(self, input_cols, search=None, replace_by=None, search_by="chars", output_cols=None):
        """
        Replace a value, list of values by a specified string
        :param input_cols: '*', list of columns names or a single column name.
        :param output_cols:
        :param search: Values to look at to be replaced
        :param replace_by: New value to replace the old one
        :param search_by: Can be "full","words","chars" or "numeric".
        :return:
        """

        # TODO check if .contains can be used instead of regexp
        def func_chars_words(_df, _input_col, _output_col, _search, _replace_by):
            # Reference https://www.oreilly.com/library/view/python-cookbook/0596001673/ch03s15.html

            # Create as dict
            _search_and_replace_by = None
            if is_list_value(search):
                _search_and_replace_by = {s: _replace_by for s in search}
            elif is_one_element(search):
                _search_and_replace_by = {search: _replace_by}

            _search_and_replace_by = {str(k): str(v) for k, v in _search_and_replace_by.items()}
            _regex = re.compile("|".join(map(re.escape, _search_and_replace_by.keys())))

            def multiple_replace(_value, __search_and_replace_by):
                print("__search_and_replace_by11111111111", __search_and_replace_by)
                # Create a regular expression from all of the dictionary keys
                if _value is not None:
                    __regex = None
                    if search_by == "chars":
                        __regex = re.compile("|".join(map(re.escape, __search_and_replace_by.keys())))
                    elif search_by == "words":
                        __regex = re.compile(
                            r'\b%s\b' % r'\b|\b'.join(map(re.escape, __search_and_replace_by.keys())))
                    result = __regex.sub(lambda match: __search_and_replace_by[match.group(0)], str(_value))
                    print("result11111111-----", result)
                else:
                    result = None

                return result

            print("_search_and_replace_by", _search_and_replace_by)
            print("multiple_replace", multiple_replace)
            return self.apply(_input_col, multiple_replace, "str", (_search_and_replace_by,), output_cols=_output_col)

        def func_full(_df, _input_col, _output_col, _search, _replace_by):
            _search = val_to_list(search)

            if _input_col != output_col:
                _df = _df.cols.copy(_input_col, _output_col)

            return _df.replace(_search, _replace_by, _output_col)

        def func_numeric(_df, _input_col, _output_col, _search, _replace_by):
            _df = _df.withColumn(_output_col,
                                 F.when(df[_input_col] == _search, _replace_by).otherwise(df[_output_col]))
            return _df

        func = None
        if search_by == "full":
            func = func_full
        elif search_by == "chars" or search_by == "words":
            func = func_chars_words
        elif search_by == "numeric":
            func = func_numeric
        else:
            RaiseIt.value_error(search_by, ["chars", "words", "full", "numeric"])

        filter_dtype = None

        df = self.root

        if search_by in ["chars", "words", "full"]:
            filter_dtype = [df.constants.STRING_TYPES]
        elif search_by == "numeric":
            filter_dtype = [df.constants.NUMERIC_TYPES]

        columns = prepare_columns(df.data, input_cols, output_cols)
        # columns = prepare_columns(df.data, input_cols, output_cols, filter_by_column_types=filter_dtype)
        dfd = df.data
        for input_col, output_col in columns:
            dfd = func(dfd, input_col, output_col, search, replace_by)

            meta = Meta.action(df.meta, None, Actions.REPLACE.value, output_col)
        return self.root.new(dfd, meta=meta)

    @staticmethod
    def replace_regex(input_cols, regex=None, value=None, output_cols=None):
        """
        Use a Regex to replace values
        :param input_cols: '*', list of columns names or a single column name.
        :param output_cols:
        :param regex: values to look at to be replaced
        :param value: new value to replace the old one
        :return:
        """

        # if regex or normal replace we use regexp or replace functions
        def func_regex(_input_cols, attr):
            _search = attr[0]
            _replace = attr[1]
            return F.regexp_replace(_input_cols, _search, _replace)

        return Cols.apply(input_cols, func=func_regex, args=[regex, value], output_cols=output_cols,
                          filter_col_by_dtypes=self.root.constants.STRING_TYPES + self.root.constants.NUMERIC_TYPES,
                          meta_action=Actions.REPLACE_REGEX.value)

    def impute(self, input_cols, data_type="continuous", strategy="mean", output_cols=None):
        """
        Imputes missing data from specified columns using the mean or median.
        :param input_cols: list of columns to be analyze.
        :param output_cols:
        :param data_type: "continuous" or "categorical"
        :param strategy: String that specifies the way of computing missing data. Can be "mean", "median" for continuous
        or "mode" for categorical columns
        :return: Dataframe object (DF with columns that has the imputed values).
        """
        df = self.root

        if data_type == "continuous":
            input_cols = parse_columns(df, input_cols,
                                       filter_by_column_types=df.constants.NUMERIC_TYPES)
            check_column_numbers(input_cols, "*")
            output_cols = get_output_cols(input_cols, output_cols)

            # Imputer require not only numeric but float or double
            # print("{} values imputed for column(s) '{}'".format(df.cols.count_na(input_col), input_col))
            df = df.cols.cast(input_cols, "float", output_cols)
            imputer = Imputer(inputCols=output_cols, outputCols=output_cols)

            model = imputer.setStrategy(strategy).fit(df)

            df = model.transform(df)

        elif data_type == "categorical":

            input_cols = parse_columns(df, input_cols,
                                       filter_by_column_types=df.constants.STRING_TYPES)
            check_column_numbers(input_cols, "*")
            output_cols = get_output_cols(input_cols, output_cols)

            value = df.cols.mode(input_cols)
            df = df.cols.fill_na(output_cols, value, output_cols)
        else:
            RaiseIt.value_error(data_type, ["continuous", "categorical"])

        return df

    def fill_na(self, input_cols, value=None, output_cols=None):
        """
        Replace null data with a specified value
        :param input_cols: '*', list of columns names or a single column name.
        :param output_cols:
        :param value: value to replace the nan/None values
        :return:
        """
        df = self.root
        input_cols = parse_columns(df, input_cols)
        check_column_numbers(input_cols, "*")
        output_cols = get_output_cols(input_cols, output_cols)

        for input_col, output_col in zip(input_cols, output_cols):
            func = None
            if is_column_a(self, input_col, df.constants.NUMERIC_TYPES):

                new_value = fastnumbers.fast_float(value)
                func = F.when(df.functions.match_nulls_strings(input_col), new_value).otherwise(F.col(input_col))
            elif is_column_a(self, input_col, df.constants.STRING_TYPES):
                new_value = str(value)
                func = F.when(df.functions.match_nulls_strings(input_col), new_value).otherwise(F.col(input_col))
            elif is_column_a(self, input_col, df.constants.ARRAY_TYPES):
                if is_one_element(value):
                    new_value = F.array(F.lit(value))
                else:
                    new_value = F.array(*[F.lit(v) for v in value])
                func = F.when(df.functions.match_null(input_col), new_value).otherwise(F.col(input_col))
            else:
                if df.cols.data_type(input_col) == parse_python_dtypes(type(value).__name__):

                    new_value = value
                    func = F.when(df.functions.match_null(input_col), new_value).otherwise(F.col(input_col))
                else:
                    RaiseIt.type_error(value, [df.cols.data_type(input_col)])

            df = df.cols.apply(input_col, func=func, output_cols=output_col, meta_action=Actions.FILL_NA.value)

        return df

    def is_na(self, input_cols, output_cols=None):
        """
        Replace null values with True and non null with False
        :param input_cols: '*', list of columns names or a single column name.
        :param output_cols:
        :return:
        """

        def _replace_na(_col_name, _value):
            return F.when(F.col(_col_name).isNull(), True).otherwise(False)

        return self.apply(input_cols, _replace_na, output_cols=output_cols, meta_action=Actions.IS_NA.value)

    @staticmethod
    def count_zeros(columns):
        """
        Count zeros in a column
        :param columns: '*', list of columns names or a single column name.
        :return:
        """
        columns = parse_columns(self.root, columns)

        return format_dict(Cols.agg_exprs(columns, self.root.functions.zeros_agg))


    @staticmethod
    def unique(columns):
        """
        Return uniques values from a columns
        :param columns:
        :return:
        """
        columns = parse_columns(self.root, columns)

        # .value(columns, "1")

        result = {}
        for col_name in columns:
            result.update(compress_list(self.select(col_name).distinct().to_dict()))
        return result


    # Stats
    @staticmethod
    def z_score(input_cols, output_cols=None):
        """
        Return the column z score
        :param input_cols: '*', list of columns names or a single column name
        :param output_cols:
        :return:
        """

        df = self.root

        def _z_score(col_name, attr):
            mean_value = df.cols.mean(col_name)
            stdev_value = df.cols.std(col_name)
            return F.abs((F.col(col_name) - mean_value) / stdev_value)

        input_cols = parse_columns(df, input_cols)

        # Hint the user if the column has not the correct data type
        for input_col in input_cols:
            if not is_column_a(df, input_col, df.constants.NUMERIC_TYPES):
                print(
                    "'{}' column is not numeric, z-score can not be calculated. Cast column to numeric using df.cols.cast()".format(
                        input_col))

        return Cols.apply(input_cols, func=_z_score, filter_col_by_dtypes=df.constants.NUMERIC_TYPES,
                          output_cols=output_cols,
                          meta_action=Actions.Z_SCORE.value)

    @staticmethod
    def standard_scaler(input_cols, output_cols=None):
        return 1

    @staticmethod
    def min_max_scaler(input_cols, output_cols=None):
        """
        Return the column min max scaler result
        :param input_cols: '*', list of columns names or a single column name
        :param output_cols:
        :return:
        """

        # Spark suuport this by default. But this implemntation support multiple columns
        def _min_max(col_name, attr):
            range_value = self.root.cols.range(col_name)
            min_value = range_value[col_name]["range"]["min"]
            max_value = range_value[col_name]["range"]["max"]
            return F.abs((F.col(col_name) - min_value) / max_value - min_value)

        return Cols.apply(input_cols, func=_min_max, filter_col_by_dtypes=self.root.constants.NUMERIC_TYPES,
                          output_cols=output_cols,
                          meta_action=Actions.MIN_MAX_SCALER.value)

    @staticmethod
    def max_abs_scaler(input_cols, output_cols=None):
        """
        Return the max abs scaler result
        :param input_cols: '*', list of columns names or a single column name
        :param output_cols:
        :return:
        """

        # Spark suuport this by default. But this implemntation support multiple columns
        def _result(col_name, attr):
            def max_abs(col_name):
                return F.max(F.abs(F.col(col_name)))

            max_abs_result = format_dict(Cols.agg_exprs(input_cols, max_abs))

            return (F.col(col_name)) / max_abs_result

        return Cols.apply(input_cols, func=_result, filter_col_by_dtypes=self.root.constants.NUMERIC_TYPES,
                          output_cols=output_cols,
                          meta_action=Actions.MAX_ABS_SCALER.value)

    @staticmethod
    def iqr(columns, more=None, relative_error=RELATIVE_ERROR):
        """
        Return the column Inter Quartile Range
        :param columns:
        :param more: Return info about q1 and q3
        :param relative_error:
        :return:
        """
        iqr_result = {}
        df = self.root
        columns = parse_columns(
            df, columns, filter_by_column_types=df.constants.NUMERIC_TYPES)
        check_column_numbers(columns, "*")

        quartile = df.cols.percentile(columns, [0.25, 0.5, 0.75], relative_error=relative_error)
        for col_name in columns:
            q1 = quartile[col_name]["percentile"]["0.25"]
            q2 = quartile[col_name]["percentile"]["0.5"]
            q3 = quartile[col_name]["percentile"]["0.75"]

            iqr_value = q3 - q1
            if more:
                result = {"iqr": iqr_value, "q1": q1, "q2": q2, "q3": q3}
            else:
                result = iqr_value
            iqr_result[col_name] = result

        return format_dict(iqr_result)

    def create_key(self, col="id") -> 'DataFrameType':
        dfd = self.data.withColumn(col, F.monotonically_increasing_id())
        return self.root.new(dfd)

    @staticmethod
    # TODO: Maybe we should create nest_to_vector and nest_array, nest_to_string
    def nest(input_cols, separator="", output_col=None, drop=True, shape="string"):
        """
        Concat multiple columns to one with the format specified
        :param input_cols: columns to be nested
        :param separator: char to be used as separator at the concat time
        :param shape: final data type, 'array', 'string' or 'vector'
        :param output_col:
        :return: Spark DataFrame
        """
        df = self.root
        output_col = parse_columns(df, output_col, accepts_missing_cols=True)
        check_column_numbers(output_col, 1)

        output_ordered_columns = df.cols.names()

        if output_col not in output_ordered_columns:
            output_ordered_columns.append(output_col)

        if has_(input_cols, F.Column):
            # Transform non Column data to lit
            input_cols = [F.lit(col) if not is_(col, F.col) else col for col in input_cols]
        else:
            input_cols = parse_columns(df, input_cols)

        if shape is "vector":
            input_cols = parse_columns(df, input_cols,
                                       filter_by_column_types=df.constants.NUMERIC_TYPES)
            output_col = one_list_to_val(output_col)
            vector_assembler = VectorAssembler(
                inputCols=input_cols,
                outputCol=output_col)

            df = vector_assembler.transform(df)

            df.meta = Meta.action(df.meta, None, Actions.NEST.value, output_col)

        elif shape is "array":
            # Arrays needs all the elements with the same data type. We try to cast to type
            df = df.cols.cast("*", "str")
            df = df.cols.apply(input_cols, F.array(*input_cols), output_cols=output_col,
                               skip_output_cols_processing=True, meta_action=Actions.NEST.value)

        elif shape is "string":
            df = df.cols.apply(input_cols, F.concat_ws(separator, *input_cols), output_cols=output_col,
                               skip_output_cols_processing=True, meta_action=Actions.NEST.value)
        else:
            RaiseIt.value_error(shape, ["vector", "array", "string"])

        df.meta = Meta.action(df.meta, None, Actions.NEST.value, output_col)

        if output_col not in output_ordered_columns:
            col_index = output_ordered_columns.index(input_cols[-1]) + 1
            output_ordered_columns[col_index:col_index] = [output_col]

        if drop is True:
            for input_col in input_cols:
                if input_col in output_ordered_columns and input_col != output_col:
                    output_ordered_columns.remove(input_col)

        return df.cols.select(output_ordered_columns)

    @staticmethod
    def unnest(input_cols, separator=None, splits=None, index=None, output_cols=None, drop=False) -> DataFrame:
        """
        Split an array or string in different columns
        :param input_cols: Columns to be un-nested
        :param output_cols: Resulted on or multiple columns  after the unnest operation [(output_col_1_1,output_col_1_2), (output_col_2_1, output_col_2]
        :param separator: char or regex
        :param splits: Number of rows to un-nested. Because we can not know beforehand the number of splits
        :param index: Return a specific index per columns. [{1,2},()]
        :param drop:
        """

        # If a number of split was not defined try to infer the length with the first element
        infer_splits = None
        if splits is None:
            infer_splits = True

        # Special case. A dot must be escaped
        if separator is not None:
            separator = re.escape(separator)

        df = self.root

        input_cols = parse_columns(df, input_cols)
        output_cols = get_output_cols(input_cols, output_cols)
        final_columns = None

        def _final_columns(_index, _splits, _output_col):

            if _index is None:
                actual_index = builtins.range(0, _splits)
            else:
                _index = val_to_list(_index)

                if is_list_of_tuples(_index):
                    _index = [(i - 1, j - 1) for i, j in _index]
                elif is_list_value(_index):
                    _index = [i - 1 for i in _index]

                actual_index = _index

            # Create final output columns
            if is_tuple(_output_col):
                columns = zip(actual_index, _output_col)
            else:
                columns = [(i, _output_col + "_" + str(i)) for i in actual_index]
            return columns

        for idx, (input_col, output_col) in enumerate(zip(input_cols, output_cols)):

            # If numeric convert and parse as string.
            if is_column_a(df, input_col, df.constants.NUMERIC_TYPES):
                df = df.cols.cast(input_col, "str")

            # Parse depending of data types
            if is_column_a(df, input_col, "struct"):
                # Unnest a data Struct
                df = df.select(output_col + ".*")

            # Array
            elif is_column_a(df, input_col, "array"):
                # Try to infer the array length using the first row
                if infer_splits is True:
                    splits = format_dict(df.agg(F.max(F.size(input_col))).to_dict())

                expr = F.col(input_col)
                final_columns = _final_columns(index, splits, output_col)
                for i, col_name in final_columns:
                    df = df.withColumn(col_name, expr.getItem(i))

            # String
            elif is_column_a(df, input_col, "str"):
                if separator is None:
                    RaiseIt.value_error(separator, "regular expression")

                # Try to infer the array length using the first row
                if infer_splits is True:
                    splits = format_dict(df.agg(F.max(F.size(F.split(F.col(input_col), separator)))).to_dict())

                expr = F.split(F.col(input_col), separator)
                final_columns = _final_columns(index, splits, output_col)
                for i, col_name in final_columns:
                    df = df.withColumn(col_name, expr.getItem(i))

            # Vector
            # TODO: Maybe we could implement Pandas UDF for better control columns output
            elif is_column_a(df, input_col, "vector"):

                def _unnest(row):
                    _dict = row.asDict()

                    # Get the column we want to unnest
                    _list = _dict[input_col]

                    # Ensure that float are python floats and not np floats
                    if index is None:
                        _list = [float(x) for x in _list]
                    else:
                        _list = [float(_list[1])]

                    return row + tuple(_list)

                df = df.rdd.map(_unnest).toDF(df.columns)

            else:
                RaiseIt.type_error(input_col, ["string", "struct", "array", "vector"])
            df.meta = Meta.action(df.meta, None, Actions.UNNEST.value, [v for k, v in final_columns])
        return df

    @staticmethod
    def heatmap(columns, buckets=10):
        """
        Returns heat map data in a dictionary
        :param columns:
        :param buckets: number of buckets
        :return:
        """

        if len(columns) != 2:
            RaiseIt.length_error(columns, "2")

        df = self.root
        columns = parse_columns(df, columns)

        values = df.cols.range(columns)

        for col_name in columns:
            # Create splits
            splits = create_buckets(values[col_name]["range"]["min"], values[col_name]["range"]["max"], buckets)

            # Create buckets in the dataFrame
            df = df.cols.bucketizer(col_name, splits=splits, output_cols=name_col(col_name, "bucketizer"))

        columns_bucket = [name_col(col_name, "bucketizer") for col_name in columns]

        size_name = "count"
        result = df.groupby(columns_bucket).agg(F.count('*').alias(size_name),
                                                F.round((F.max(columns[0]) + F.min(columns[0])) / 2).alias(
                                                    columns[0]),
                                                F.round((F.max(columns[1]) + F.min(columns[1])) / 2).alias(
                                                    columns[1]),
                                                ).rows.sort(columns).toPandas()
        x = result[columns[0]].tolist()
        y = result[columns[1]].tolist()
        s = result[size_name].tolist()

        return {"x": {"name": columns[0], "data": x}, "y": {"name": columns[1], "data": y}, "s": s}

    def hist(self, columns="*", buckets=20, compute=True):
        df = self.root
        # return df.cols.agg_exprs(columns, self.F.min, compute=compute, tidy=True)

        result = df.cols.agg_exprs(columns, self.F.hist_agg, buckets, None, df=self.root)

        # TODO: for some reason casting to int in the exprs do not work. Casting Here. A Spark bug?
        # Example
        # Column < b'array(map(count, CAST(sum(CASE WHEN ((rank >= 7) AND (rank < 7.75)) THEN 1 ELSE 0 END) AS INT),
        # lower, 7, upper, 7.75) AS `hist_agg_rank_0`, map(count, CAST(sum(CASE WHEN ((rank >= 7.75) AND (rank < 8.5))
        # THEN 1 ELSE 0 END) AS INT), lower, 7.75, upper, 8.5) AS `hist_agg_rank_1`, map(count,
        # CAST(sum(CASE WHEN ((rank >= 8.5) AND (rank < 9.25)) THEN 1 ELSE 0 END) AS INT), lower, 8.5, upper, 9.25)
        # AS `hist_agg_rank_2`, map(count, CAST(sum(CASE WHEN ((rank >= 9.25) AND (rank < 10))
        # THEN 1 ELSE 0 END) AS INT), lower, 9.25, upper, 10) AS `hist_agg_rank_3`) AS `histrank`' >

        return {"hist": result}

        # TODO: In tests this code run faster than using Cols.agg_exprs when run over all the columns.
        #  Not when running over columns individually
        # columns = parse_columns(self, columns)
        # df = self
        # for col_name in columns:
        #     print(Cols.agg_exprs(hist_agg, col_name, self, buckets))
        #     # print(df.agg(hist_agg(col_name, self, buckets)))
        # return result

    def frequency(self, columns="*", n=10, percentage=False, total_rows=None, count_uniques=False, compute=True):
        """
        Output values frequency in json format
        :param columns: Columns to be processed
        :param n: n top elements
        :param percentage: Get
        :param total_rows: Total rows to calculate the percentage. If not provided is calculated
        :return:
        """
        df = self.root
        columns = parse_columns(df, columns)

        dfd = df.data
        if columns is not None:

            # Convert non compatible columns(non str, int or float) to string
            non_compatible_columns = self.names(columns)

            if non_compatible_columns is not None:
                dfd = self.root.cols.cast(non_compatible_columns, "str").data

            freq = (dfd.select(columns).rdd
                    .flatMap(lambda x: x.asDict().items())
                    .map(lambda x: (x, 1))
                    .reduceByKey(lambda a, b: a + b)
                    .groupBy(lambda x: x[0][0])
                    .flatMap(lambda g: nlargest(n, g[1], key=lambda x: x[1]))
                    .repartition(1)  # Because here we have small data move all to 1 partition
                    .map(lambda x: (x[0][0], (x[0][1], x[1])))
                    .groupByKey().map(lambda x: (x[0], list(x[1]))))

            result = {}
            for f in freq.collect():
                result[f[0]] = {"count_uniques": "N/A", "values": [{"value": kv[0], "count": kv[1]} for kv in f[1]]}

            # if count_uniques:
            #     print(dfd.count())

            if percentage:
                if total_rows is None:
                    total_rows = dfd.count()

                    RaiseIt.type_error(total_rows, ["int"])
                for col_name in columns:
                    for c in result[col_name]:
                        c["percentage"] = round((c["count"] * 100 / total_rows), 2)

            return {"frequency": result}

    def correlation(self, input_cols, method="pearson", output="json"):
        """
        Calculate the correlation between columns. It will try to cast a column to float where necessary and impute
        missing values
        :param input_cols: Columns to be processed
        :param method: Method used to calculate the correlation
        :param output: array or json
        :return:
        """

        df = self.root

        # Values in columns can not be null. Warn user
        input_cols = parse_columns(
            df, input_cols, filter_by_column_types=df.constants.NUMERIC_TYPES)

        # Input is not a vector transform to a vector
        output_col = name_col(input_cols, "correlation")
        check_column_numbers(input_cols, ">1")

        for col_name in input_cols:
            df = df.cols.cast(col_name, "float")
            logger.print("Casting {col_name} to float...".format(col_name=col_name))

        df = df.cols.nest(input_cols, "vector", output_col=output_col)

        # Correlation can not handle null values. Check if exist ans warn the user.
        cols_na = {x: y for x, y in df.cols.count_na(input_cols).items() if y != 0}

        if cols_na:
            message = "Correlation can not handle nulls. " + " and ".join(
                {str(x) + " has " + str(y) + " null(s)" for x, y in cols_na.items()})
            RaiseIt.message(ValueError, message)

        corr = Correlation.corr(df, output_col, method).head()[0].toArray()
        result = None
        if output is "array":
            result = corr

        elif output is "json":
            # Parse result to json
            col_pair = []
            for col_name in input_cols:
                for col_name_2 in input_cols:
                    col_pair.append({"between": col_name, "an": col_name_2})

            # flat array
            values = corr.flatten('F').tolist()

            result = []
            for n, v in zip(col_pair, values):
                # Remove correlation between the same column
                if n["between"] is not n["an"]:
                    n["value"] = v
                    result.append(n)

            result = sorted(result, key=lambda k: k['value'], reverse=True)

        return {"cols": input_cols, "data": result}

    def schema_data_type(self, columns="*"):
        """
        Return the column(s) data type as Type
        :param columns: Columns to be processed
        :return:
        """
        df = self.root
        columns = parse_columns(df, columns)
        return format_dict([df.data.schema[col_name].dataType for col_name in columns])



    def qcut(self, columns, quantiles, handle_invalid="skip"):
        """
        Bin columns into n buckets. Quantile Discretizer
        :param columns: Input columns to processed
        :param num_buckets: Number of buckets in which the column will be divided
        :param handle_invalid:
        :return:
        """
        df = self.root
        columns = parse_columns(
            df, columns, filter_by_column_types=df.constants.NUMERIC_TYPES)
        check_column_numbers(columns, "*")

        for col_name in columns:
            discretizer = QuantileDiscretizer(numBuckets=quantiles, inputCol=col_name,
                                              outputCol=name_col(col_name, "qcut"),
                                              handleInvalid=handle_invalid)
            df = discretizer.fit(df).transform(df)
        return df

    @staticmethod
    def clip(columns, lower_bound, upper_bound):
        """
        Trim values at input thresholds
        :param columns: Columns to be trimmed
        :param lower_bound: Lower value bound
        :param upper_bound: Upper value bound
        :return:
        """

        df = self.root

        columns = parse_columns(
            df, columns, filter_by_column_types=df.constants.NUMERIC_TYPES)
        check_column_numbers(columns, "*")

        def _clip(_col_name, args):
            _lower = args[0]
            _upper = args[1]
            return (F.when(F.col(_col_name) <= _lower, _lower)
                    .when(F.col(_col_name) >= _upper, _upper)).otherwise(F.col(col_name))

        for col_name in columns:
            df = df.cols.apply_expr(col_name, _clip, [lower_bound, upper_bound])
        return df

    def string_to_index(self, cols=None, output_cols=None):
        """
        Encodes a string column of labels to a column of label indices
        :param cols:
        :param output_cols:
        :param columns:
        :return:
        """
        df = self.root

        df = ml_string_to_index(df, cols, output_cols)

        return df

    def index_to_string(self, cols=None, output_cols=None):
        """
        Encodes a string column of labels to a column of label indices
        :param cols:
        :param output_cols:
        :param columns:
        :return:
        """
        df = self.root

        df = ml_index_to_string(df, cols, output_cols)

        return df

    @staticmethod
    def bucketizer(input_cols, splits, output_cols=None):
        """
        Bucketize multiples columns at the same time.
        :param input_cols:
        :param splits: Dict of splits or ints. You can use create_buckets() to make it
        :param output_cols:
        :return:
        """
        df = self.root

        if is_int(splits):
            min_max = df.cols.range(input_cols)[input_cols]["range"]
            splits = create_buckets(min_max["min"], min_max["max"], splits)

        def _bucketizer(col_name, args):
            """
            Create a column expression that create buckets in a range of values
            :param col_name: Column to be processed
            :return:
            """

            buckets = args
            expr = []

            for i, b in enumerate(buckets):
                if i == 0:
                    expr = when((F.col(col_name) >= b["lower"]) & (F.col(col_name) <= b["upper"]), b["bucket"])
                else:
                    expr = expr.when((F.col(col_name) >= b["lower"]) & (F.col(col_name) <= b["upper"]),
                                     b["bucket"])

            return expr

        df = df.cols.apply(input_cols, func=_bucketizer, args=splits, output_cols=output_cols)
        return df



import numpy as np
from pprint import pformat
from abc import abstractmethod, ABC
import operator
import time

import humanize
import imgkit
import jinja2
import simplejson as json
from glom import assign
from tabulate import tabulate

from optimus.helpers.types import *
from optimus.helpers.check import is_notebook
from optimus.helpers.columns import parse_columns
from optimus.helpers.constants import BUFFER_SIZE, Actions, RELATIVE_ERROR
from optimus.helpers.core import val_to_list
from optimus.helpers.functions import df_dicts_equal, absolute_path, reduce_mem_usage, update_dict
from optimus.helpers.json import json_converter
from optimus.helpers.output import print_html
from optimus.helpers.converter import convert_numpy
from optimus.infer import is_str, is_tuple, is_list
from optimus.profiler.constants import MAX_BUCKETS
from optimus.profiler.templates.html import HEADER, FOOTER
from optimus.engines.base.columns import *
from optimus.engines.base.rows import *
from optimus.engines.base.mask import Mask
from optimus.engines.base.set import BaseSet
from optimus.engines.base.profile import BaseProfile
from optimus.engines.base.meta import Meta
from optimus.engines.base.ml.models import BaseML
from optimus.engines.base.ml.encoding import BaseEncoding
from optimus.engines.base.io.save import *
from optimus.engines.base.constants import BaseConstants
from optimus.engines.base.functions import BaseFunctions
from optimus.outliers.outliers import Outliers
from optimus.plots.functions import plot_hist, plot_frequency
from optimus.plots.plots import Plot


class BaseDataFrame(ABC):
    """
    Optimus DataFrame
    """

    def __init__(self, data: 'InternalDataFrameType', op: 'EngineType', label_encoder=None):
        data = self._compatible_data(data)
        self.data = data
        self.buffer = None
        self.updated = None
        self.meta = {}
        self.le = label_encoder
        self.op = op

        # .profile and .set are properties to support docstrings
        self.profile = BaseProfile(self)
        self.set = BaseSet(self)

    @staticmethod
    def _compatible_data(data):
        return data

    def __del__(self):
        del self.data
        del self.le

    @property
    def root(self) -> 'DataFrameType':
        return self

    def _repr_html_(self):
        df = self
        return df.table()

    def __repr__(self):
        df = self
        return df.ascii()

    def __getitem__(self, item):

        if isinstance(item, slice):
            return self.buffer_window("*", item.start, item.stop)
        elif is_str(item) or is_list(item):
            return self.cols.select(item)
        elif isinstance(item, BaseDataFrame):
            return self.rows.select(item)

    def __setitem__(self, key, value):

        df = self.cols.assign({key: value})
        self.data = df.data
        self.buffer = df.buffer
        self.updated = df.updated
        self.meta = df.meta
        self.le = df.le
        self.op = df.op

    def __len__(self):
        return self.rows.count()

    def new(self, dfd, meta=None) -> 'DataFrameType':
        df = self.__class__(dfd, op=self.op)
        if meta is not None:
            df.meta = meta
        import copy
        df.le = copy.deepcopy(self.le)
        return df

    def copy(self) -> 'DataFrameType':
        """
        Return a copy of a dataframe
        """
        df = self.root
        import copy
        return self.root.new(df.data.copy(), meta=copy.deepcopy(df.meta))

    @staticmethod
    def __operator__(df, data_type=None, multiple_columns=False) -> 'DataFrameType':
        if isinstance(df, (BaseDataFrame,)):
            col1 = "*" if multiple_columns else df.cols.names(0)[0]

            if data_type:
                df = df.cols.cast(col1, data_type).data
            else:
                df = df.data

            if not multiple_columns:
                df = df[col1]

        return df

    @abstractmethod
    def _base_to_dfd(self, pdf, n_partitions) -> 'InternalDataFrameType':
        pass

    @abstractmethod
    def to_optimus_pandas(self) -> 'DataFrameType':
        pass

    def unary_operation(self, df, opb, data_type=None) -> 'DataFrameType':
        """
        Helper to process binary operations
        :param df: Left
        :param opb: Operator
        :param data_type: 
        :return:
        """
        df = BaseDataFrame.__operator__(df, data_type, True)

        return self.new(opb(df))

    def operation(self, df1, df2, opb, data_type=None) -> 'DataFrameType':
        """
        Helper to process binary operations
        :param df1: Left
        :param df2: Right
        :param opb: Operator
        :param data_type:
        :return:
        """

        if isinstance(df1, (np.generic,)):
            df1 = np.asscalar(df1)
        
        if isinstance(df2, (np.generic,)):
            df2 = np.asscalar(df2)

        if is_list(df1):
            df1 = self.op.create.dataframe({"0": df1})
        if is_list(df2):
            df2 = self.op.create.dataframe({"0": df2})
        
        df1_is_df = isinstance(df1, (BaseDataFrame,))
        df2_is_df = isinstance(df2, (BaseDataFrame,))

        if not df1_is_df or not df2_is_df:
            multiple_columns = True
            if data_type == "auto":
                # finds the type of the value
                not_df = df2 if df1_is_df else df1
                data_type = type(not_df).__name__
                data_type = "float" if data_type == "int" else data_type
            elif data_type in ["float", "int"]:
                df1 = df1 if df1_is_df else float(df1)
                df2 = df2 if df2_is_df else float(df2)
            elif data_type == "str":
                df1 = df1 if df1_is_df else str(df1)
                df2 = df2 if df2_is_df else str(df2)


        else:
            multiple_columns = df1.cols.names() == df2.cols.names()
            if data_type == "auto":
                data_type = None

        df1 = BaseDataFrame.__operator__(df1, data_type, multiple_columns)
        df2 = BaseDataFrame.__operator__(df2, data_type, multiple_columns)

        if multiple_columns:
            dfd = opb(df1, df2)

        else:
            name_left = name_right = ""

            if not isinstance(df1, (int, float, str, dict, list)):
                name_left = getattr(df1, "name", 0)

            if not isinstance(df2, (int, float, str, dict, list)):
                name_right = getattr(df2, "name", 0)

            if name_left and name_right:
                name = (name_left + "_" +
                        name_right) if name_left != name_right else name_left
            else:
                name = name_left if name_left else name_right

            dfd = opb(df1, df2).rename(name).to_frame()

        return self.new(dfd)

    def __invert__(self) -> 'DataFrameType':
        return self.unary_operation(self, operator.invert)

    def __neg__(self) -> 'DataFrameType':
        return self.unary_operation(self, operator.neg)

    def __add__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.add, "auto")

    def __radd__(self, df2) -> 'DataFrameType':
        return self.operation(df2, self, operator.add, "auto")

    def __sub__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.sub, "float")

    def __rsub__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.sub, "float")

    def __mul__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.mul, "float")

    def __rmul__(self, df2) -> 'DataFrameType':
        return self.operation(df2, self, operator.mul, "float")

    def __truediv__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.truediv, "float")

    def __rtruediv__(self, df2) -> 'DataFrameType':
        return self.operation(df2, self, operator.truediv, "float")

    def __floordiv__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.floordiv, "float")

    def __rfloordiv__(self, df2) -> 'DataFrameType':
        return self.operation(df2, self, operator.floordiv, "float")

    def __mod__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.mod, "float")

    def __rmod__(self, df2) -> 'DataFrameType':
        return self.operation(df2, self, operator.mod, "float")

    def __pow__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.pow, "float")

    def __rpow__(self, df2) -> 'DataFrameType':
        return self.operation(df2, self, operator.pow, "float")

    def __eq__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.eq)

    def __gt__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.gt, "float")

    def __lt__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.lt, "float")

    def __ne__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.ne)

    def __ge__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.ge, "float")

    def __le__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.le, "float")

    def __and__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.__and__, "bool")

    def __or__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.__or__, "bool")

    def __xor__(self, df2) -> 'DataFrameType':
        return self.operation(self, df2, operator.__xor__, "bool")

    def _to_values(self):
        """
        Return values from a dataframe in numpy or cupy format. Aimed to be used internally in Machine Learning models
        :return:
        """
        return self.data.values

    def equals_dataframe(self, df2: 'DataFrameType') -> bool:
        df2 = df2.data
        return self.data.equals(df2)

    def equals(self, df2: 'DataFrameType', decimal=None, assertion=False) -> bool:
        df2_is_dataframe = isinstance(df2, (BaseDataFrame,))

        # checks by column names
        if df2_is_dataframe:
            cols2 = df2.cols.names()
        else:
            cols2 = list(df2.keys())

        cols1 = self.cols.names()
        
        if cols1 != cols2:
            if assertion:
                raise AssertionError(f"Column names are not equal: {cols1}, {cols2}")
            return False

        if decimal is not None or assertion:
            # checks by each column
            if df2_is_dataframe:
                df2 = df2.to_dict(n="all")
            df1 = self.to_dict(n="all")

            return df_dicts_equal(df1, df2, decimal=decimal, assertion=assertion)

        else:
            # checks by dataframe
            if df2_is_dataframe:
                result = self.equals_dataframe(df2)
                if not result and assertion:
                    raise AssertionError("Dataframes are not equal")
                return result
            else:
                return df_dicts_equal(self.to_dict(n="all"), df2, decimal=decimal, assertion=assertion)

    @property
    @abstractmethod
    def save(self) -> 'BaseSave':
        pass

    @property
    def functions(self) -> 'BaseFunctions':
        return BaseFunctions(self)

    @property
    def mask(self) -> 'Mask':
        return Mask(self)

    @property
    def ml(self) -> 'BaseML':
        return BaseML(self)

    @property
    @abstractmethod
    def rows(self) -> 'BaseRows':
        pass

    @property
    @abstractmethod
    def cols(self) -> 'BaseColumns':
        pass

    @property
    def constants(self) -> 'BaseConstants':
        return BaseConstants()

    @property
    def plot(self) -> 'Plot':
        return Plot(self)

    @property
    def outliers(self) -> 'Outliers':
        return Outliers(self)

    @property
    def encoding(self) -> 'BaseEncoding':
        return BaseEncoding(self)

    def visualize(self):
        raise NotImplementedError(f"\"visualize\" is not available using {type(self).__name__}")

    def execute(self) -> 'DataFrameType':
        return self

    def compute(self) -> 'InternalDataFrameType':
        return self.data

    def _assign(self, kw_columns):

        dfd = self.data
        kw_columns = {str(key): kw_column for key, kw_column in kw_columns.items()}
        return dfd.assign(**kw_columns)

    def to_json(self, cols="*", n="all", orient="list") -> str:
        """

        :param cols:
        :param n:
        :param orient:
        The format of the JSON string:

        ‘split’ : dict like {‘index’ -> [index], ‘columns’ -> [columns], ‘data’ -> [values]}
        ‘records’ : list like [{column -> value}, … , {column -> value}]
        ‘index’ : dict like {index -> {column -> value}}
        ‘columns’ : dict like {column -> {index -> value}}
        ‘values’ : just the values array
        ‘table’ : dict like {‘schema’: {schema}, ‘data’: {data}}

        Describing the data, where data component is like orient='records'.
        :return:
        """


        return json.dumps(self.to_dict(cols, n, orient), ensure_ascii=False, default=json_converter)

    def to_dict(self, cols="*", n=10, orient="list") -> dict:
        """
            Return a dict from a Collect result
            :param cols:
            :param n:
            :param orient:
            :return:
        """
        if n == "all":
            dfd = self.cols.select(cols).to_pandas()
        else:
            dfd = self.buffer_window(cols, 0, n).data

        return dfd.to_dict(orient)

    @staticmethod
    @abstractmethod
    def sample(n=10, random=False):
        pass

    def columns_sample(self, cols="*") -> dict:
        """
        Return a dict of the sample of a Dataframe
        :return:
        """
        df = self

        cols = parse_columns(df, cols)        

        return {"columns": [{"title": col_name} for col_name in cols],
                "value": df.rows.to_list(cols)}

    @abstractmethod
    def to_pandas(self):
        pass

    def stratified_sample(self, col_name, seed: int = 1) -> 'DataFrameType':
        """
        Stratified Sampling
        :param col_name:
        :param seed:
        :return:
        """
        raise NotImplementedError(f"\"stratified_sample\" is not available using {type(self).__name__}")

    def get_buffer(self) -> 'DataFrameType':
        """
        Get buffer from the dataframe
        """
        return self.buffer

    @abstractmethod
    def _buffer_window(self, input_cols, lower_bound, upper_bound):
        pass

    def _reset_buffer(self):
        self.buffer = None

    def reset_buffer(self):
        self._reset_buffer()
        self.meta = Meta.reset(self.meta, "buffer_time")

    def buffer_window(self, cols="*", lower_bound=None, upper_bound=None, n=BUFFER_SIZE):
        """
        Get a window from the buffer of the dataframe
        """
        df = self

        meta = df.meta
        buffer_time = Meta.get(meta, "buffer_time")
        last_action_time = Meta.get(meta, "last_action_time")

        if buffer_time and last_action_time:
            if buffer_time > last_action_time:
                self.set_buffer(cols, n)

        if lower_bound is None or lower_bound < 0:
            lower_bound = 0

        input_columns = parse_columns(df, cols)
        return self._buffer_window(input_columns, lower_bound, upper_bound)

    def buffer_json(self, columns):
        df = self.buffer
        columns = parse_columns(df, columns)

        return {"columns": [{"title": col_name} for col_name in df.cols.select(columns).cols.names()],
                "value": df.rows.to_list(columns)}

    def size(self, deep=True, format=None):
        """
        Get the size of the dataframe in bytes
        :return:
        """
        df = self.data
        result = df.memory_usage(index=True, deep=deep).sum()
        if format == "human":
            result = humanize.naturalsize(result)

        return result

    def optimize(self, categorical_threshold=50, verbose=False):
        df = self
        return reduce_mem_usage(df, categorical_threshold=categorical_threshold, verbose=verbose)

    def run(self):
        """
        This method is a very useful function to break lineage of transformations. By default Spark uses the lazy
        evaluation approach in processing data: transformation functions are not computed into an action is called.
        Sometimes when transformations are numerous, the computations are very extensive because the high number of
        operations that spark needs to run in order to get the results.

        Other important thing is that Apache Spark save task but not result of dataFrame, so tasks are
        accumulated and the same situation happens.

        :return:
        """
        df = self
        df.cache().count()
        return df

    @staticmethod
    @abstractmethod
    def query(sql_expression):
        raise NotImplementedError

    @staticmethod
    def is_cached(df):
        """

        :return:
        """
        return False if Meta.get(df.meta, "profile") is None else True

    def _cols_to_profile(self, columns):
        """
        Get the columns that needs to be profiled and renames the columns in the metadata
        :return:
        """

        df = self
        actions = Meta.get(df.meta, "transformations.actions")
        has_actions = actions is not None and len(actions) > 0

        profiler_columns = Meta.get(df.meta, "profile.columns")
        if profiler_columns is not None:
            profiler_columns = {
                col_name: value for col_name, value in profiler_columns.items() if value.get("data_type", None)
            }

        new_columns = parse_columns(df, columns) or []

        if profiler_columns is None:
            calculate_columns = new_columns
        else:
            profiled_columns = list(profiler_columns.keys())
            if not has_actions:
                calculate_columns = [
                    column for column in new_columns if column not in profiled_columns]

            else:
                modified_columns = []
                # Operations need to be processed int the same order that created
                for action in Meta.get(df.meta, "transformations.actions"):
                    dropped_columns = []
                    
                    action_name = action.get("name", "action")
                    column = action.get("columns", None)

                    if is_tuple(column):
                        source, target = column

                    else:
                        source = target = column

                    if action_name == Actions.COPY.value and source in profiler_columns:
                        profiler_columns[target] = profiler_columns[source]

                    elif action_name == Actions.RENAME.value and source in profiler_columns:
                        profiler_columns[target] = profiler_columns[source]
                        profiler_columns.pop(source)
                        if source in new_columns:
                            new_columns.pop(source)
                            new_columns.push(target)

                    elif action_name == Actions.DROP.value and source in profiler_columns:
                        profiler_columns.pop(source)
                        dropped_columns.append(source)

                    else:
                        modified_columns.append(source)

                profiled_columns = list(profiler_columns.keys())

                calculate_columns = [
                    column for column in new_columns if column not in profiled_columns]
                modified_columns = [
                    column for column in modified_columns if column in new_columns]
                calculate_columns = list(
                    set(modified_columns + calculate_columns))
                calculate_columns = list(
                    set(calculate_columns) - set(dropped_columns))

        return calculate_columns

    @abstractmethod
    def partitions(self):
        pass

    @staticmethod
    def partitioner():
        raise NotImplementedError

    def repartition(self, n=None, *args, **kwargs):
        df = self.data
        return self.new(df, meta=self.meta)

    def table_image(self, path, limit=10):
        """
        Output table as image
        :param limit:
        :param path:
        :return:
        """

        css = absolute_path("/css/styles.css")

        imgkit.from_string(self.table_html(
            limit=limit, full=True), path, css=css)
        print_html("<img src='" + path + "'>")

    def table_html(self, limit=10, cols=None, title=None, full=False, truncate=True, count=True, highlight=[]):
        """
        Return a HTML table with the spark cols, data types and values
        :param cols: Columns to be printed
        :param limit: How many rows will be printed
        :param title: Table title
        :param full: Include html header and footer
        :param truncate: Truncate the row information
        :param count:

        :return:
        """

        cols = parse_columns(self, cols)
        if limit is None:
            limit = 10

        df = self

        total_rows = df.rows.approx_count()

        if limit == "all":
            limit = total_rows
            data = df.cols.select(cols).to_dict(n="all", orient="records")
        else:
            limit = min(limit, total_rows)
            data = df.cols.select(cols).rows.limit(
                limit + 1).to_dict(n="all", orient="records")
        # Load the Jinja template
        template_loader = jinja2.FileSystemLoader(
            searchpath=absolute_path("/templates/out"))
        template_env = jinja2.Environment(
            loader=template_loader, autoescape=True)
        template = template_env.get_template("table.html")

        # Filter only the columns and data type info need it
        data_types = [(k, v) for k, v in df.cols.data_type(tidy=False)["data_type"].items()]

        # Remove not selected columns
        final_columns = []
        for i in data_types:
            for j in cols:
                if i[0] == j:
                    final_columns.append(i)

        total_rows = humanize.intword(total_rows)
        total_cols = df.cols.count()
        total_partitions = df.partitions()
        df_type = type(df)
        highlight = val_to_list(highlight)
        output = template.render(df_type=df_type, cols=final_columns, data=data, limit=limit, total_rows=total_rows,
                                 total_cols=total_cols,
                                 partitions=total_partitions, title=title, truncate=truncate, highlight=highlight)

        if full is True:
            output = HEADER + output + FOOTER
        return output

    def display(self, limit=10, cols=None, title=None, truncate=True, plain_text=False, highlight=None):
        """

        :param limit:
        :param cols:
        :param title:
        :param truncate:
        :param plain_text:
        :param highlight:
        :return:
        """
        if highlight is None:
            highlight = []
        df = self

        if is_notebook() and not plain_text:
            print_html(df.table(limit, cols, title, truncate, highlight))

        else:
            print(df.ascii(limit, cols))

    def print(self, limit=10, cols=None):
        print(self.ascii(limit, cols))

    def table(self, limit=None, cols=None, title=None, truncate=True, highlight=None):
        """
        Print a dataframe in html format
        :param limit: The number of files that will be printed
        :param cols: Select the columns to be printed
        :param title:
        :param truncate:
        :param highlight:
        :return:
        """
        if highlight is None:
            highlight = []
        df = self
        try:
            if is_notebook():
                # TODO: move the html param to the ::: if is_notebook() and engine.output is "html":
                return df.table_html(title=title, limit=limit, cols=cols, truncate=truncate, highlight=highlight)

        except NameError as e:
            print(e)

        return df.ascii(limit, cols)

    def ascii(self, limit=10, cols=None):
        """
        Print a dataframe in ascii format
        :param limit:
        :param cols:
        :return:
        """
        df = self
        if not cols:
            cols = "*"

        limit = min(limit, df.rows.approx_count())
        return tabulate(df.rows.limit(limit + 1).cols.select(cols).to_pandas(),
                        headers=[f"""{i}\n({j})""" for i,
                                 j in df.cols.data_type(tidy=False)["data_type"].items()],
                        tablefmt="simple",
                        showindex="never") + "\n"

    def export(self, n="all", data_types="inferred"):
        """
        Helper function to export all the dataframe in dict format. Aimed to be used in test functions
        :return:
        """
        df_dict = self.to_dict(n=n)

        if not data_types:
            df_data = pformat(df_dict, sort_dicts=False,
                              width=800, compact=True)
        else:
            if data_types == "internal":
                df_dtypes = self.cols.data_type(tidy=False)["data_type"]
            else:
                df_dtypes = self.cols.infer_type(tidy=False)["infer_type"]
                df_dtypes = {col: df_dtypes[col]["data_type"] for col in df_dtypes}
            df_data = []
            for col_name in df_dict.keys():
                value = pformat((col_name, df_dtypes[col_name]))
                value += ": "
                value += pformat(df_dict[col_name],
                                 sort_dicts=False, width=800, compact=True)
                df_data.append(value)

            df_data = "{" + ", ".join(df_data) + "}"

        return df_data

    def show(self, n=10):
        """
        :return:
        """
        return self.data.head(n=n)

    @staticmethod
    @abstractmethod
    def debug():
        pass

    def reset(self):
        # df = self.df
        df = self
        df.meta = {}
        return df

    def calculate_profile(self, cols="*", bins: int = MAX_BUCKETS, flush: bool = False, size=False):
        """
        Returns a new dataframe with the profile data in its added to the meta property
        :param columns:
        :param bins:
        :param flush:
        :param size: get the dataframe size in memory. Use with caution this could be slow for big data frames.
        :return:
        """
        _t = time.process_time()
        profiler_time = {"hist": {}, "frequency": {}, "count_mismatch": {}}

        df = self
        meta = self.meta

        if flush is False:
            cols_to_profile = df._cols_to_profile(cols) or []
        else:
            cols_to_profile = parse_columns(df, cols) or []

        profiler_data = Meta.get(meta, "profile")

        is_cached = profiler_data is not None

        if profiler_data is None:
            profiler_data = {}
        cols_data_types = None

        profiler_time["beginning"] = {"elapsed_time": time.process_time() - _t}

        if cols_to_profile or not is_cached or flush:

            if flush:
                meta = Meta.set(meta, "profile", {})
                df.meta = meta

            hist_cols = []
            freq_cols = []

            cols_data_types = {}
            cols_to_infer = [*cols_to_profile]

            for col_name in cols_to_profile:
                col_data_type = Meta.get(df.meta, f"columns_data_types.{col_name}")

                if col_data_type is not None:
                    cols_data_types[col_name] = col_data_type
                    cols_to_infer.remove(col_name)

            if cols_to_infer:
                cols_data_types = {**cols_data_types, **df.cols.infer_type(cols_to_infer, tidy=False)["infer_type"]}
                cols_data_types = {col: cols_data_types[col] for col in cols_to_profile if col in cols_data_types}

            _t = time.process_time()
            mismatch = df.cols.quality(cols_data_types)
            profiler_time["count_mismatch"] = {
                "columns": cols_data_types, "elapsed_time": time.process_time() - _t}

            # Get with columns are numerical and does not have mismatch so we can calculate the histogram
            cols_properties = cols_data_types.items()
            for col_name, properties in cols_properties:
                if properties.get("data_type") in df.constants.NUMERIC_TYPES \
                   and not properties.get("categorical", False):
                    hist_cols.append(col_name)
                else:
                    freq_cols.append(col_name)

            hist = None
            freq = {}
            sliced_freq = {}
            count_uniques = None

            if len(hist_cols):
                _t = time.process_time()
                hist = df.cols.hist(hist_cols, buckets=bins, compute=False)
                profiler_time["hist"] = {
                    "columns": hist_cols, "elapsed_time": time.process_time() - _t}

            if len(freq_cols):
                _t = time.process_time()
                sliced_cols = []
                non_sliced_cols = []

                # Extract the columns with cells larger thatn
                max_cell_length = getattr(df.meta, "max_cell_length", None)

                if max_cell_length:
                    for i, j in max_cell_length.items():
                        if i in freq_cols:
                            if j > 50:
                                sliced_cols.append(i)
                            else:
                                non_sliced_cols.append(i)

                else:
                    non_sliced_cols = freq_cols

                if len(non_sliced_cols) > 0:
                    # print("non_sliced_cols",non_sliced_cols)
                    freq = df.cols.frequency(
                        non_sliced_cols, n=bins, count_uniques=True, compute=False)

                if len(sliced_cols) > 0:
                    # print("sliced_cols", sliced_cols)
                    sliced_freq = df.cols.slice(sliced_cols, 0, 50).cols.frequency(sliced_cols, n=bins,
                                                                                   count_uniques=True,
                                                                                   compute=False)

                profiler_time["frequency"] = {
                    "columns": freq_cols, "elapsed_time": time.process_time() - _t}

            def merge(_columns, _hist, _freq, _mismatch, _data_types, _count_uniques):
                _c = {}

                _hist = {} if _hist is None else _hist.get("hist", {})
                _freq = {} if _freq is None else _freq.get("frequency", {})

                for _col_name in _columns:
                    _c[_col_name] = {
                        "stats": _mismatch.get(_col_name, None),
                        "data_type": _data_types.get(_col_name, None)
                    }
                    if _col_name in _freq:
                        f = _freq[_col_name]
                        _c[_col_name]["stats"]["frequency"] = f["values"]
                        _c[_col_name]["stats"]["count_uniques"] = f["count_uniques"]

                    elif _col_name in _hist:
                        h = _hist[_col_name]
                        _c[_col_name]["stats"]["hist"] = h

                return {"columns": _c}

            # Nulls
            total_count_na = 0

            data_types = df.cols.data_type("*", tidy=False)["data_type"]

            hist, freq, sliced_freq, mismatch = self.functions.compute(
                hist, freq, sliced_freq, mismatch)

            freq = {**freq, **sliced_freq}

            updated_columns = merge(
                cols_to_profile, hist, freq, mismatch, data_types, count_uniques)

            for col in list(profiler_data.get("columns", {}).keys()):
                if col in updated_columns["columns"]:
                    del profiler_data["columns"][col]

            profiler_data = update_dict(profiler_data, updated_columns)

            assign(profiler_data, "name", Meta.get(df.meta, "name"), dict)
            assign(profiler_data, "file_name",
                   Meta.get(df.meta, "file_name"), dict)

            data_set_info = {'cols_count': df.cols.count(),
                             'rows_count': df.rows.count(),
                             }
            if size is True:
                data_set_info.update({'size': df.size(format="human")})

            assign(profiler_data, "summary", data_set_info, dict)

            data_types_list = list(set(df.cols.data_type("*", tidy=False)["data_type"].values()))

            assign(profiler_data, "summary.data_types_list", data_types_list, dict)
            assign(profiler_data, "summary.total_count_data_types",
                   len(set([i for i in data_types.values()])), dict)
            assign(profiler_data, "summary.missing_count", total_count_na, dict)
            
            rows_count = df.rows.count()

            if rows_count:
                assign(profiler_data, "summary.p_missing", round(
                    total_count_na / rows_count * 100, 2))
            else:
                assign(profiler_data, "summary.p_missing", None)

        # _t = time.process_time()

        all_columns_names = df.cols.names()

        # meta = Meta.set(meta, "transformations", value={})

        # Order columns
        actual_columns = profiler_data["columns"]
        profiler_data["columns"] = {key: actual_columns[key]
                                    for key in all_columns_names if key in actual_columns}
        meta = Meta.set(meta, "profile", profiler_data)

        if cols_data_types is not None:
            df.meta = meta
            df = df.cols.set_data_type(cols_data_types, inferred=True)
            meta = df.meta

        # Reset Actions
        meta = Meta.reset_actions(meta, parse_columns(df, cols or []))
        df.meta = meta
        profiler_time["end"] = {"elapsed_time": time.process_time() - _t}
        # print(profiler_time)
        return df

    def graph(self) -> dict:
        raise NotImplementedError(f"Not supported using {type(self).__name__}")

    def get_series(self):
        col1 = self.cols.names(0)[0]
        return self.data[col1]

    def join(self, df_right: 'DataFrameType', how="left", on=None, left_on=None, right_on=None, key_middle=False) -> 'DataFrameType':
        """
        Join 2 dataframes SQL style
        :param df_right:
        :param how{‘left’, ‘right’, ‘outer’, ‘inner’, ‘exclusive’, ‘exclusive left’, ‘exclusive right’}, default ‘left’
        :param on:
        :param left_on:
        :param right_on:
        :param key_middle: Order the columns putting the left df columns before the key column and the right df columns

        :return:
        """
        # if not is_(df_right, BaseDataFrame):
        #     RaiseIt.type_error(df_right, ["BaseDataFrame"])

        suffix_left = "_left"
        suffix_right = "_right"

        df_left = self.root

        if on is not None:
            left_on = on
            right_on = on

        if df_left.cols.data_type(left_on) == "category":
            df_left[left_on] = df_left[left_on].cat.as_ordered()

        if df_right.cols.data_type(right_on) == "category":
            df_right[right_on] = df_right[right_on].cat.as_ordered()

        # Join does not work with different data types.
        df_left[left_on] = df_left[left_on].cols.cast("*", "str")
        df_left.data.set_index(left_on)

        df_right[right_on] = df_right[right_on].cols.cast("*", "str")
        df_right.data.set_index(right_on)

        # Used to reorder the output
        left_names = df_left.cols.names()
        right_names = df_right.cols.names()


        if how in ['exclusive', 'exclusive left', 'exclusive right']:
            _how = 'outer'
            indicator = True
        else:
            _how = how
            indicator = False

        dfd = df_left.data.merge(df_right.data, how=_how, left_on=left_on,
                                 right_on=right_on, suffixes=(suffix_left, suffix_right),
                                 indicator=indicator)

        if how == 'exclusive':
            dfd = dfd[(dfd["_merge"] == "left_only") | (dfd["_merge"] == "right_only")]
        elif how == 'exclusive left':
            dfd = dfd[dfd["_merge"] == "left_only"]
        elif how == 'exclusive right':
            dfd = dfd[dfd["_merge"] == "right_only"]

        if indicator:
            dfd = dfd.drop(["_merge"], axis=1)

        df = self.root.new(dfd)

        # Reorder
        last_column_name = left_names[-1]
        if key_middle is True:
            names = df.cols.names()
            last_column_name = last_column_name if last_column_name in names else last_column_name + suffix_left
            left_on = left_on if left_on in names else left_on + suffix_left
            right_on = right_on if right_on in names else right_on + suffix_right
            if left_on in names:
                df = df.cols.move(left_on, "before", last_column_name)
            if right_on in names:
                df = df.cols.move(right_on, "before", last_column_name)

        return df

    def string_clustering(self, cols="*", algorithm="fingerrint", *args, **kwargs):
        from optimus.engines.base.stringclustering import string_clustering
        return string_clustering(self, cols, algorithm, *args, **kwargs)
        # return clusters

    def agg(self, aggregations: dict, groupby=None, output="dict", tidy=True):
        """
        :param aggregations: Dictionary or list of tuples with the form [("col", "agg")]
        :param groupby: None, list of columns names or a single column name to group the aggregations.
        :param output{‘dict’, ‘dataframe’}, default ‘dict’: Output type.
        """

        df = self
        dfd = df.data

        if is_dict(aggregations):
            aggregations = aggregations.items()

        if groupby:
            groupby = parse_columns(df, groupby)

            for column, aggregations_set in aggregations:
                aggregations[column] = val_to_list(aggregations_set)

            dfd = dfd.groupby(groupby).agg(aggregations)

            dfd.columns = ['_'.join(col).strip() for col in dfd.columns.values]

            if output == "dict":
                result = dfd.to_dict()

            elif output == "dataframe":
                dfd.columns = [str(c) for c in dfd.columns]
                result = self.new(dfd.reset_index())

        else:
            result = {}

            for column, aggregations_set in aggregations:
                aggregations_set = val_to_list(aggregations_set)
                for aggregation in aggregations_set:
                    result[column + "_" + aggregation] = getattr(
                        df.cols, aggregation)(column, tidy=True)

            if output == "dataframe":
                result = self.op.create.dataframe({k: [v] for k, v in result.items()})

        return convert_numpy(format_dict(result, tidy=tidy))

    def report(self, df, cols="*", buckets=MAX_BUCKETS, infer=False, relative_error=RELATIVE_ERROR,
               approx_count=True,
               mismatch=None, advanced_stats=True):
        """
        Return dataframe statistical information in HTML Format
        :param df: Dataframe to be analyzed
        :param columns: Columns to be analyzed
        :param buckets: Number of buckets calculated to print the histogram
        :param infer: infer data type
        :param relative_error: Relative Error for quantile discretizer calculation
        :param approx_count: Use approx_count_distinct or countDistinct
        :param mismatch:
        :param advanced_stats:
        :return:
        """

        cols = parse_columns(df, cols)
        output = self.dataset(df, cols, buckets, infer, relative_error, approx_count, format="dict",
                              mismatch=mismatch, advanced_stats=advanced_stats)

        # Load jinja
        template_loader = jinja2.FileSystemLoader(
            searchpath=absolute_path("/profiler/templates/out"))
        template_env = jinja2.Environment(
            loader=template_loader, autoescape=True)

        # Render template
        # Create the profiler info header
        html = ""
        general_template = template_env.get_template("general_info.html")
        html = html + general_template.render(data=output)

        template = template_env.get_template("one_column.html")
        # Create every column stats
        for col_name in cols:
            hist_pic = None
            freq_pic = None

            col = output["columns"][col_name]
            if "hist" in col["stats"]:
                hist_dict = col["stats"]["hist"]

                if col["column_data_type"] == "date":
                    hist_year = plot_hist(
                        {col_name: hist_dict["years"]}, "base64", "years")
                    hist_month = plot_hist(
                        {col_name: hist_dict["months"]}, "base64", "months")
                    hist_weekday = plot_hist(
                        {col_name: hist_dict["weekdays"]}, "base64", "weekdays")
                    hist_hour = plot_hist(
                        {col_name: hist_dict["hours"]}, "base64", "hours")
                    hist_minute = plot_hist(
                        {col_name: hist_dict["minutes"]}, "base64", "minutes")
                    hist_pic = {"hist_years": hist_year, "hist_months": hist_month, "hist_weekdays": hist_weekday,
                                "hist_hours": hist_hour, "hist_minutes": hist_minute}

                elif col["column_data_type"] == "int" or col["column_data_type"] == "string" or col[
                        "column_data_type"] == "float":
                    hist = plot_hist({col_name: hist_dict}, output="base64")
                    hist_pic = {"hist_numeric_string": hist}
            if "frequency" in col:
                freq_pic = plot_frequency(
                    {col_name: col["frequency"]}, output="base64")

            html = html + \
                template.render(data=col, freq_pic=freq_pic, hist_pic=hist_pic)

        # Save in case we want to output to a html file
        # self.html = html + df.table_html(10)
        self.html = html

        # Display HTML
        print_html(self.html)

        # JSON
        # Save in case we want to output to a json file
        self.json = output

        return self



from abc import abstractmethod

import dask
import dask.array as da
import dask.dataframe as dd
import humanize
import numpy as np
import pandas as pd
from distributed import Variable
from dask.utils import parse_bytes

from optimus.engines.base.basedataframe import BaseDataFrame
from optimus.engines.base.distributed.dataframe import DistributedBaseDataFrame
from optimus.engines.pandas.dataframe import PandasDataFrame
from optimus.helpers.core import val_to_list
from optimus.helpers.functions import random_int
from optimus.helpers.raiseit import RaiseIt
from optimus.helpers.types import *
from optimus.infer import is_int


class DaskBaseDataFrame(DistributedBaseDataFrame):

    def _assign(self, kw_columns: dict):

        dfd = self.root.data

        for key in kw_columns:
            kw_column = kw_columns[key]

            if isinstance(kw_column, (list,)):
                kw_column = pd.Series(kw_column)

            if isinstance(kw_column, pd.Series):
                # TO-DO: A Pandas Series should be assignable to a Dask DataFrame
                kw_column = dd.from_pandas(kw_column, npartitions=dfd.npartitions)

            if isinstance(kw_column, (np.ndarray, da.Array)):
                kw_column = dd.from_array(kw_column)

            if isinstance(kw_column, (dd.Series, dd.DataFrame)):

                kw_column.index = dfd.index

                if isinstance(kw_column, dd.DataFrame):
                    if key in kw_column:
                        # the incoming series has the same column key
                        kw_column = kw_column[key]
                    else:
                        # the incoming series has no column key
                        kw_column = kw_column[list(kw_column.columns)[0]]
                else:
                    kw_column.name = key

            kw_columns[key] = kw_column

        kw_columns = {str(key): kw_column for key, kw_column in kw_columns.items()}
        return dfd.assign(**kw_columns)

    @staticmethod
    @abstractmethod
    def _base_to_dfd(df, n_partitions):
        """
        Convert a dataframe from the dataframe base (pandas, cudf) in a distributed engine dataframe (dask, dask_cudf)
        :param pdf:
        :param n_partitions:
        :return:
        """
        pass

    def execute(self):
        self.data = self.data.persist()
        return self

    def compute(self):
        df = self.data
        return df.compute()

    def visualize(self):
        return display(self.data.visualize())

    def _reset_buffer(self):
        if self.buffer:
            Variable(self.buffer).delete()
            self.buffer = None

    def get_buffer(self):
        return Variable(self.buffer).get() if self.buffer else None

    def _buffer_window(self, input_cols, lower_bound, upper_bound):
        def func(value):
            return value[lower_bound:upper_bound].reset_index(drop=True)

        return PandasDataFrame(self.data[input_cols].partitions[0].map_partitions(func).compute(), op=self.op,
                               label_encoder=self.le)

    def graph(self) -> dict:
        """
        Return a dict the Dask tasks graph
        :return:
        """
        return self.data.__dask_graph__().layers

    def sample(self, n=10, random=False):
        """
        Return a n number of sample from a dataFrame
        :param n: Number of samples
        :param random: if true get a semi random sample
        :return:
        """
        if not is_int(n):
            RaiseIt.type_error(n, ["int"])

        if random is True:
            seed = int(random_int())
        elif random is False:
            seed = 0
        elif is_int(random):
            seed = random
        else:
            RaiseIt.type_error(random, ["bool", "int"])

        df = self.root
        rows_count = df.rows.count()
        if n < rows_count:
            # n/rows_count can return a number that represent less the total number we expect. multiply by 1.001
            fraction = (n / rows_count) * 1.001
        else:
            fraction = 1.0

        dfd = df.data.sample(frac=fraction, random_state=seed).reset_index(drop=True)

        return self.root.new(dfd)

    def stratified_sample(self, col_name, seed: int = 1):
        """
        Stratified Sampling
        :param col_name:
        :param seed:
        :return:
        """
        df = self.data
        n = min(5, df[col_name].value_counts().min())
        df = df.groupby(col_name).apply(lambda x: x.sample(2))
        # df_.index = df_.index.droplevel(0)
        return self.root.new(df)

    def pivot(self, col, groupby, agg=None, values=None):
        """
        Return reshaped DataFrame organized by given index / column values.
        :param col: Column to use to make new frame's columns.
        :param groupby: Column to use to make new frame's index.
        :param agg: Aggregation to use for populating new frame's values.
        :param values: Column to use for populating new frame's values.
        :return:
        """

        df = self.root

        if values is not None:
            agg = ("first", values)

        groupby = val_to_list(groupby)
        by = groupby + [col]
        if agg is None:
            agg = ("count", col)
        return df.cols.groupby(by=by, agg=agg).unstack(by)

    @staticmethod
    def melt(id_vars, value_vars, var_name="variable", value_name="value", data_type="str"):
        """
        Convert DataFrame from wide to long format.
        :param id_vars: column with unique values
        :param value_vars: Column names that are going to be converted to columns values
        :param var_name: Column name for vars
        :param value_name: Column name for values
        :param data_type: All columns must have the same type. It will transform all columns to this data type.
        :return:
        """

        raise NotImplementedError

    def size(self, deep=False, format=None):
        """
        Get the size of a dask in bytes
        :return:
        """
        df = self.data
        result = df.memory_usage(index=True, deep=deep).sum().compute()
        if format == "human":
            result = humanize.naturalsize(result)

        return result

    def run(self):
        """
        This method is a very useful function to break lineage of transformations. By default Spark uses the lazy
        evaluation approach in processing data: transformation functions are not computed into an action is called.
        Sometimes when transformations are numerous, the computations are very extensive because the high number of
        operations that spark needs to run in order to get the results.

        Other important thing is that Apache Spark save task but not result of dataFrame, so tasks are
        accumulated and the same situation happens.

        :return:
        """
        df = self.data
        df.cache().count()
        return df

    @staticmethod
    def query(sql_expression):
        raise NotImplementedError

    def partitions(self):
        return self.data.npartitions

    def equals_dataframe(self, df2: 'DataFrameType') -> bool:
        if isinstance(df2, (BaseDataFrame,)):
            return df2.to_dict(n="all") == self.to_dict(n="all")
        else:
            return self.to_dict(n="all") == df2

    @staticmethod
    def partitioner():
        print("Dask not support custom partitioner")
        raise NotImplementedError

    def repartition(self, n=None, *args, **kwargs):
        dfd = self.data
        df = self
        if n == "auto":
            # Follow a heuristic for partitioning a mentioned
            # https://docs.dask.org/en/latest/best-practices.html#avoid-very-large-partitions
            client = distributed.get_client()
            worker_memory = parse_bytes(client.cluster.worker_spec[0]["options"]["memory_limit"])
            nthreads = client.cluster.worker_spec[0]["options"]["nthreads"]

            part_recommended_size = worker_memory / nthreads / 10
            n = int(df.size() / part_recommended_size)

            # Partition can not be lower than 1
            n = n if n < 0 else 1
            # TODO .repartition(partition_size="100MB"). https://stackoverflow.com/questions/44657631/strategy-for-partitioning-dask-dataframes-efficiently
        dfd = dfd.repartition(npartitions=n, *args, **kwargs)

        return self.new(dfd, meta=self.meta)

    @staticmethod
    def debug():
        """

        :return:
        """
        raise NotImplementedError

    def to_dict(self, cols="*", n=10, orient="list"):
        """
        Create a dict
        :param n:
        :param orient:
        :return:
        """

        if n == "all":
            series = self.cols.select(cols).to_pandas()
        else:
            series = self.buffer_window(cols, 0, n).data

        return series.to_dict(orient)

    def to_pandas(self):
        return self.data.compute()

    @property
    def constants(self):
        from optimus.engines.base.dask.constants import Constants
        return Constants()



from dask_ml.cluster import KMeans
from dask_ml.linear_model import LinearRegression, LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn.externals import joblib

from optimus.engines.base.ml.models import BaseML
from optimus.helpers.columns import parse_columns
from optimus.helpers.types import *


class ML(BaseML):
    def __init__(self, root: 'DataFrameType'):
        super().__init__(root)
        self.root = root

    def linear_regression(self, features, target, *args, **kwargs):
        """
        Fit a linear regression model. This ensure that the data is ready converting to numeric and dropping NA
        :param features:
        :param target:
        :param args:
        :param kwargs:
        :return:
        """
        df = self.root
        lm = LinearRegression(*args, **kwargs)

        features = parse_columns(df, features)
        target = parse_columns(df, target)

        df = df.cols.select(features + target).cols.to_float().rows.drop_missings()
        X_train = df[features]._to_values()
        y_train = df[target]._to_values()

        return lm.fit(X_train, y_train)

    def logistic_regression(self, features, target, *args, **kwargs):
        df = self.root
        lm = LogisticRegression(*args, **kwargs)

        features = parse_columns(df, features)
        target = parse_columns(df, target)

        df = df.cols.select(features + target).cols.to_float().rows.drop_missings()
        # LogisticRegression requires that the Dask Array passed to it has known chunk sizes
        # https://stackoverflow.com/questions/61756328/dusk-ml-logisticregression-throws-this-error-notimplementederror-can-not-add
        X_train = df[features]._to_values().compute_chunk_sizes()
        y_train = df[target]._to_values().compute_chunk_sizes()

        return lm.fit(X_train, y_train)

    def k_means(self, features, n_centers, *args, **kwargs):
        df = self.root
        k_means = KMeans(n_clusters=n_centers)
        features = parse_columns(df, features)
        df = df.cols.select(features).cols.to_float().rows.drop_missings()
        X_train = df[features]._to_values().compute_chunk_sizes()
        print(X_train)
        k_means.fit(X_train)
        return

    def random_forest(self, features, target, *args, **kwargs):
        """
        Runs a random forest classifier for input DataFrame.
        :return: Model with random forest and prediction run.
        """
        df = self.root
        da_rf = RandomForestClassifier(*args, **kwargs)

        df = df.cols.select(features + target).cols.to_float().rows.drop_missings()
        # LogisticRegression requires that the Dask Array passed to it has known chunk sizes
        # https://stackoverflow.com/questions/61756328/dusk-ml-logisticregression-throws-this-error-notimplementederror-can-not-add
        X = df[features]._to_values().compute_chunk_sizes()
        y = df[target]._to_values().compute_chunk_sizes()

        with joblib.parallel_backend("dask"):
            da_rf.fit(X, y)
        return da_rf



import re
from abc import abstractmethod, ABC

import hidateinfer
import jellyfish
import numpy as np
import pandas as pd
from fastnumbers import fast_float, fast_int
from jsonschema._format import is_email
from metaphone import doublemetaphone

from optimus.helpers.logger import logger
from optimus.helpers.constants import ProfilerDataTypes
from optimus.helpers.core import one_list_to_val, one_tuple_to_val, val_to_list
from optimus.infer import is_datetime_str, is_list, is_list_of_list, is_null, is_bool, \
    is_credit_card_number, is_zip_code, is_float_like, is_datetime, is_valid_datetime_format, \
    is_object_value, is_ip, is_url, is_missing, is_gender, is_list_of_int, is_list_of_str, \
    is_str, is_phone_number, is_int_like


# ^(?:(?P<protocol>[\w\d]+)(?:\:\/\/))?(?P<sub_domain>(?P<www>(?:www)?)(?:\.?)(?:(?:[\w\d-]+|\.)*?)?)(?:\.?)(?P<domain>[^./]+(?=\.))\.(?P<top_domain>com(?![^/|:?#]))?(?P<port>(:)(\d+))?(?P<path>(?P<dir>\/(?:[^/\r\n]+(?:/))+)?(?:\/?)(?P<file>[^?#\r\n]+)?)?(?:\#(?P<fragment>[^#?\r\n]*))?(?:\?(?P<query>.*(?=$)))*$

class BaseFunctions(ABC):
    """
    Functions for internal use or to be called using 'F': `from optimus.functions import F`
    Note: some methods needs to be static so they can be passed to a Dask worker.
    """

    _engine = None
    """The engine to be used for internal use:
        `pd` from `import pandas as pd` or `dask` from `import dask`
    """
    _partition_engine = None
    """(optional, defaults to `_engine`) The engine to be used for internal use
        in a distributed engine (if supported):
        `pd` from `import pandas as pd` which is used by `Dask`
    """

    def __init_subclass__(cls):

        if cls._partition_engine is None:
            cls._partition_engine = cls._engine

    def __init__(self, df=None):

        if self._engine is None:
            raise TypeError("Can't instantiate class without attribute '_engine'")

        if df is not None and getattr(df, "partitions", False):
            self.root = df
        else:
            self.root = None

    def __getattr__(self, name):
        type_msg = "" if self.root is None else f" using {type(self.root).__name__}"
        raise NotImplementedError(f"\"{name}\" is not available" + type_msg)

    @property
    def _functions(self):
        """
        Gets the set of functions available in the engine
        """
        return self._partition_engine

    @staticmethod
    def _format_to_dict(_s):
        if hasattr(_s, "to_dict"):
            return _s.to_dict()
        else:
            return _s

    @staticmethod
    def _format_to_list(_s):
        if hasattr(_s, "tolist"):
            return _s.tolist()
        else:
            return _s

    @staticmethod
    def _format_to_list_one(_s):
        if hasattr(_s, "tolist"):
            return one_list_to_val(_s.tolist())
        else:
            return _s

    @property
    def n_partitions(self):
        if self.root is None:
            return 1
        return self.root.partitions()

    @staticmethod
    def delayed(func):
        """
        Convert function to a delayed function (Dummy method on unsupported engines)

        :param func:
        :return:
        """
        return func

    def from_delayed(self, delayed):
        """
        Convert delayed objects to a DataFrame or Series (Dummy method on unsupported engines)

        :param delayed:
        :return:
        """
        return delayed[0]

    def to_delayed(self, delayed):
        """
        Convert DataFrame or Series to a list of delayed objects (Dummy method on unsupported engines)

        :param delayed:
        :return:
        """
        return [delayed]

    def apply_delayed(self, series, func, *args, **kwargs):
        """
        Applies a function to a Series for every partition using apply

        :param series:
        :param func:
        :param args:
        :param kwargs:
        :return:
        """
        result = self.to_delayed(series)
        result = [partition.apply(func, *args, **kwargs) for partition in result]
        result = self.from_delayed(result)
        result.index = series.index
        return result

    def map_delayed(self, series, func, *args, **kwargs):
        """
        Applies a function to a Series for every partition using map

        :param series:
        :param func:
        :param args:
        :param kwargs:
        :return:
        """
        result = self.to_delayed(series)
        result = [partition.map(func, *args, **kwargs) for partition in result]
        result = self.from_delayed(result)
        result.index = series.index
        return result

    @property
    def constants(self):
        """

        :return:
        """
        from optimus.engines.base.constants import BaseConstants
        return BaseConstants()

    @staticmethod
    def sort_df(dfd, cols, ascending):
        """
        Sort rows taking into account one column

        :param dfd:
        :param cols:
        :param ascending:
        :return:
        """
        return dfd.sort_values(cols, ascending=ascending).reset_index(drop=True)

    @staticmethod
    def reverse_df(dfd):
        """
        Reverse rows
        """
        return dfd[::-1]

    @staticmethod
    def append(dfd, dfd2):
        """
        Append two dataframes
        """
        return dfd.append(dfd2)

    def _new_series(self, *args, **kwargs):
        """
        Creates a new series (also known as column)
        """
        return self._functions.Series(*args, **kwargs)

    def compute(self, *args, **kwargs):
        return one_tuple_to_val((*(a for a in args), *(kwargs[k] for k in kwargs)))

    @staticmethod
    def to_dict(series) -> dict:
        """
        Convert series to a Python dictionary
        """
        return series.to_dict()

    @staticmethod
    def to_items(series) -> dict:
        """
        Convert series to a list of tuples [(index, value), ...]
        """
        df = series.reset_index()
        return df.to_dict(orient='split')['data']

    def _to_boolean(self, series):
        """
        Converts series to bool
        """
        return series.map(lambda v: bool(v), na_action=None).astype('bool')

    def to_boolean(self, series):
        return self._to_boolean(series)

    def _to_boolean_none(self, series):
        """
        Converts series to boolean
        """
        return series.map(lambda v: bool(v), na_action='ignore').astype('object')

    def to_boolean_none(self, series):
        return self._to_boolean_none(series)

    def _to_float(self, series):
        """
        Converts a series values to floats
        """
        try:
            return self._new_series(np.vectorize(fast_float)(series, default=np.nan).flatten())
        except:
            return self._new_series(self._functions.to_numeric(series, errors='coerce')).astype('float')

    def to_float(self, series):
        """
        Converts a series values to floats
        """
        return self._to_float(series)

    def _to_integer(self, series, default=0):
        """
        Converts a series values to integers
        """
        try:
            return self._new_series(np.vectorize(fast_int)(series, default=default).flatten())
        except:
            return self._new_series(self._functions.to_numeric(series, errors='coerce').fillna(default)).astype('int')

    def to_integer(self, series, default=0):
        """
        Converts a series values to integers
        """
        return self._to_integer(series, default)

    def _to_datetime(self, series, format):
        """
        Converts a series values to datetimes
        """
        return series

    def to_datetime(self, series, format):
        """
        Converts a series values to datetimes
        """
        return self._to_datetime(series, format)

    @staticmethod
    def to_string(series):
        """
        Converts a series values to strings
        """
        return series.astype(str)

    @staticmethod
    def to_string_accessor(series):
        """
        Converts a series values to strings and returns it's main functions
        accessor
        """
        return series.astype(str).str

    @staticmethod
    def duplicated(dfd, keep, subset):
        """
        Mark the duplicated values in a DataFrame using a subset of columns
        """
        return dfd.duplicated(keep=keep, subset=subset)

    def impute(self, series, strategy, fill_value):
        """
        Impute missing values in a series
        """
        from sklearn.impute import SimpleImputer
        imputer = SimpleImputer(strategy=strategy, fill_value=fill_value)
        series_fit = series.dropna()
        if str(series.dtype) in self.constants.OBJECT_TYPES:
            series_fit = series_fit.astype(str)
        values = series_fit.values.reshape(-1, 1)
        if len(values):
            imputer.fit(values)
            return imputer.transform(series.fillna(np.nan).values.reshape(-1, 1))
        else:
            logger.warn("list to fit imputer is empty, try cols.fill_na instead.")
            return series

    # Aggregation
    def date_format(self, series):
        """
        Get the date format of a series
        """
        dtype = str(series.dtype)
        if dtype in self.constants.STRING_TYPES:
            series = series.astype(str)
            result = hidateinfer.infer(self.compute(series).values)
            if not is_valid_datetime_format(result) or True:
                result_series = self.date_formats(series)
                result = result_series.mode().head(1)[0]

                if not is_valid_datetime_format(result):
                    return False

            return result
        elif dtype in self.constants.DATETIME_INTERNAL_TYPES:
            return True

        return False

    def min(self, series, numeric=False, string=False):
        """
        Get the minimum value of a series
        """
        if numeric:
            series = self.to_float(series)

        if string or str(series.dtype) in self.constants.STRING_TYPES:
            return self.to_string(series.dropna()).min()
        else:
            return series.min()

    def max(self, series, numeric=False, string=False):
        """
        Get the maximum value of a series
        """
        if numeric:
            series = self.to_float(series)

        if string or str(series.dtype) in self.constants.STRING_TYPES:
            return self.to_string(series.dropna()).max()
        else:
            return series.max()

    def mean(self, series):
        """
        Get the mean value of a series
        """
        return self.to_float(series).mean()

    def mode(self, series):
        """
        Get the modal value of a series
        """
        if str(series.dtype) not in self.constants.NUMERIC_INTERNAL_TYPES:
            return self.delayed(self._format_to_list_one)(self.to_string(series.dropna()).mode())
        else:
            return self.delayed(self._format_to_list_one)(series.mode())

    @staticmethod
    def crosstab(series, other):
        """
        Compute a cross tabulation of two series
        """
        return pd.crosstab(series, other)

    def std(self, series):
        """
        Get the standard deviation of a series
        """
        return self.to_float(series).std()

    def sum(self, series):
        """
        Get the sum of a series
        """
        return self.to_float(series).sum()

    def cumsum(self, series):
        """
        Get the cumulative sum of a series
        """
        return self.to_float(series).cumsum()

    def cumprod(self, series):
        """
        Get the cumulative product of a series
        """
        return self.to_float(series).cumprod()

    def cummax(self, series):
        """
        Get the cumulative maximum of a series
        """
        return self.to_float(series).cummax()

    def cummin(self, series):
        """
        Get the cumulative minimum of a series
        """
        return self.to_float(series).cummin()

    def var(self, series):
        """
        Get the variance of a series
        """
        return self.to_float(series).var()

    def count_uniques(self, series, estimate=False):
        """
        Get the count of unique values in a series
        """
        return self.to_string(series).nunique()

    def unique_values(self, series, estimate=False):
        """
        Get the unique values in a series
        """
        # Cudf can not handle null so we fill it with non zero values.
        return self.to_string(series).unique()

    @staticmethod
    def count_na(series):
        """
        Get the count of missing values in a series
        """
        return series.isna().sum()
        # return {"count_na": {col_name:  for col_name in columns}}
        # return np.count_nonzero(_df[_serie].isnull().values.ravel())
        # return cp.count_nonzero(_df[_serie].isnull().values.ravel())

    @staticmethod
    @abstractmethod
    def count_zeros(series, *args):
        pass

    @staticmethod
    @abstractmethod
    def kurtosis(series):
        pass

    @staticmethod
    @abstractmethod
    def skew(series):
        pass

    # import dask.dataframe as dd
    def mad(self, series, error=False, more=False, estimate=False):

        _series = self.to_float(series).dropna()

        if not estimate:
            _series = self.compute(_series)

        if not len(_series):
            return np.nan
        else:
            median_value = _series.quantile(0.5)
            mad_value = (_series - median_value).abs().quantile(0.5)
            if more:
                return {"mad": mad_value, "median": median_value}
            else:
                return mad_value

    # TODO: dask seems more efficient triggering multiple .min() task, one for every column
    # cudf seems to be calculate faster in on pass using df.min()
    def range(self, series):
        series = self.to_float(series)
        return {"min": series.min(), "max": series.max()}

    def percentile(self, series, values=0.5, error=False, estimate=False):

        _series = self.to_float(series).dropna()

        if not estimate:
            _series = self.compute(_series)

        if not len(_series):
            return np.nan
        else:
            return self.delayed(self._format_to_dict)(_series.quantile(values))

    # def radians(series):
    #     return series.to_float().radians()
    #
    # def degrees(series, *args):
    #     return call(series, method_name="degrees")

    ###########################

    def z_score(self, series):
        t = self.to_float(series)
        return t - t.mean() / t.std(ddof=0)

    def modified_z_score(self, series, estimate):
        series = self.to_float(series)

        _series = series.dropna()

        if not estimate:
            _series = self.compute(_series)

        if not len(_series):
            return np.nan
        else:
            median = _series.quantile(0.5)
            mad = (_series - median).abs().quantile(0.5)

            return abs(0.6745 * (series - median) / mad)

    def clip(self, series, lower_bound, upper_bound):
        return self.to_float(series).clip(float(lower_bound), float(upper_bound))

    def cut(self, series, bins, labels, default):
        if is_list_of_int(bins):
            return pd.cut(self.to_float(series), bins, include_lowest=True, labels=labels)
        elif is_list_of_str(bins):
            conditions = [series.str.contains(i) for i in bins]

            return np.select(conditions, labels, default=default)

    def qcut(self, series, quantiles):
        return pd.qcut(series, quantiles)

    def abs(self, series):
        return self.to_float(series).abs()

    def exp(self, series):
        return self.to_float(series).exp()

    def lemmatize_verbs(self, series):
        import nltk
        w_tokenizer = nltk.tokenize.WhitespaceTokenizer()
        lemmatizer = nltk.stem.WordNetLemmatizer()

        def lemmatize_verbs_map(text):
            return " ".join([lemmatizer.lemmatize(w, "v") for w in w_tokenizer.tokenize(text)])

        return self.map_delayed(self.to_string(series), lemmatize_verbs_map, na_action=None)

    def word_tokenize(self, series):
        import nltk
        w_tokenizer = nltk.tokenize.WhitespaceTokenizer()

        def lemmatize_verbs_map(text):
            return w_tokenizer.tokenize(text)

        return self.map_delayed(self.to_string(series), lemmatize_verbs_map, na_action=None, meta="object")

    def word_count(self, series):
        return self.word_tokenize(series).str.len()

    def len(self, series):
        return self.to_string_accessor(series).len()

    @staticmethod
    @abstractmethod
    def sqrt(series):
        pass

    def mod(self, series, other):
        return self.to_float(series).mod(other)

    def round(self, series, decimals):
        return self.to_float(series).round(decimals)

    def pow(self, series, exponent):
        return self.to_float(series).pow(exponent)

    def floor(self, series):
        return self.to_float(series).floor()

    # def trunc(self):
    #     series = self.series
    #     return series.to_float().truncate()

    @staticmethod
    @abstractmethod
    def radians(series):
        pass

    @staticmethod
    @abstractmethod
    def degrees(series):
        pass

    @staticmethod
    @abstractmethod
    def ln(series):
        pass

    @staticmethod
    @abstractmethod
    def log(series, base):
        pass

    @staticmethod
    @abstractmethod
    def ceil(series):
        pass

    @abstractmethod
    def sin(self, series):
        pass

    @abstractmethod
    def cos(self, series):
        pass

    @abstractmethod
    def tan(self, series):
        pass

    @abstractmethod
    def asin(self, series):
        pass

    @abstractmethod
    def acos(self, series):
        pass

    @abstractmethod
    def atan(self, series):
        pass

    @abstractmethod
    def sinh(self, series):
        pass

    @abstractmethod
    def cosh(self, series):
        pass

    @abstractmethod
    def tanh(self, series):
        pass

    @abstractmethod
    def asinh(self, series):
        pass

    @abstractmethod
    def acosh(self, series):
        pass

    @abstractmethod
    def atanh(self, series):
        pass

    # Strings
    def match(self, series, regex):
        return self.to_string_accessor(series).match(regex)

    def lower(self, series):
        return self.to_string_accessor(series).lower()

    def upper(self, series):
        return self.to_string_accessor(series).upper()

    def title(self, series):
        return self.to_string_accessor(series).title()

    def capitalize(self, series):
        return self.to_string_accessor(series).capitalize()

    def pad(self, series, width, side, fillchar=""):
        return self.to_string_accessor(series).pad(width, side, fillchar)

    def extract(self, series, regex):
        return self.to_string_accessor(series).extract(regex)

    def slice(self, series, start, stop, step):
        return self.to_string_accessor(series).slice(start, stop, step)

    def trim(self, series):
        return self.to_string_accessor(series).strip()

    def strip_html(self, series):
        return self.to_string(series).replace('<.*?>', '', regex=True)

    def replace_chars(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)
        if ignore_case:
            str_regex = [r'(?i)%s' % re.escape(s) for s in search]
        else:
            str_regex = [r'%s' % re.escape(s) for s in search]
        return self.to_string(series).replace(str_regex, replace_by, regex=True)

    def replace_words(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)
        if ignore_case:
            str_regex = [r'(?i)\b%s\b' % re.escape(s) for s in search]
        else:
            str_regex = [r'\b%s\b' % re.escape(s) for s in search]
        return self.to_string(series).replace(str_regex, replace_by, regex=True)

    def replace_full(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)
        if ignore_case:
            str_regex = [r'(?i)^%s$' % re.escape(s) for s in search]
        else:
            str_regex = [r'^%s$' % re.escape(s) for s in search]
        return series.replace(str_regex, replace_by, regex=True)

    def replace_values(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)

        if ignore_case:
            regex = True
            search = [(r'(?i)%s' % re.escape(s)) for s in search]
        else:
            regex = False

        if is_list(replace_by) and is_list_of_list(search):
            for _s, _r in zip(search, replace_by):
                series = series.replace(_s, _r, regex=regex)

        else:
            series = series.replace(search, replace_by, regex=regex)

        return series

    def replace_regex_chars(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)
        if ignore_case:
            str_regex = [r'(?i)%s' % s for s in search]
        else:
            str_regex = [r'%s' % s for s in search]
        return self.to_string(series).replace(str_regex, replace_by, regex=True)

    def replace_regex_words(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)
        if ignore_case:
            str_regex = [r'(?i)\b%s\b' % s for s in search]
        else:
            str_regex = [r'\b%s\b' % s for s in search]
        return self.to_string(series).replace(str_regex, replace_by, regex=True)

    def replace_regex_full(self, series, search, replace_by, ignore_case):
        search = val_to_list(search, convert_tuple=True)
        if ignore_case:
            str_regex = [r'(?i)^%s$' % s for s in search]
        else:
            str_regex = [r'^%s$' % s for s in search]
        return series.replace(str_regex, replace_by, regex=True)

    def remove_numbers(self, series):
        return self.to_string_accessor(series).replace(r'\d+', '', regex=True)

    def remove_white_spaces(self, series):
        return self.to_string_accessor(series).replace(" ", "")

    def remove_urls(self, series):
        return self.to_string_accessor(series).replace(r"https?://\S+|www\.\S+", "", regex=True)

    def normalize_spaces(self, series):
        return self.to_string_accessor(series).replace(r" +", " ", regex=True)

    def expand_contractions(self, series):

        pass

    # @staticmethod
    # def len(series):
    #     return series.str.len()

    def normalize_chars(self, series):
        pass

    def find(self, sub, start=0, end=None):
        series = self.series
        return self.to_string_accessor(series).find(sub, start, end)

    def rfind(self, series, sub, start=0, end=None):
        return self.to_string_accessor(series).rfind(sub, start, end)

    def left(self, series, position):
        return self.to_string_accessor(series)[:position]

    def right(self, series, position):
        return self.to_string_accessor(series)[-1 * position:]

    def mid(self, series, _start, _n):
        return self.to_string_accessor(series)[_start:_n]

    def starts_with(self, series, pat):
        return self.to_string_accessor(series).startswith(pat)

    def ends_with(self, series, pat):
        return self.to_string_accessor(series).endswith(pat)

    def contains(self, series, pat):
        return self.to_string_accessor(series).contains(pat)

    def char(self, series, _n):
        return self.to_string_accessor(series)[_n]

    @staticmethod
    def unicode(series):
        pass

    @staticmethod
    def exact(series, pat):
        return series == pat

    # dates
    def year(self, series, format):
        """
        Extract the year from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.year

    def month(self, series, format):
        """
        Extract the month from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.month

    def day(self, series, format):
        """
        Extract the day from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.day

    def hour(self, series, format):
        """
        Extract the hour from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.hour

    def minute(self, series, format):
        """
        Extract the minute from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.minute

    def second(self, series, format):
        """
        Extract the second from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.second

    def weekday(self, series, format):
        """
        Extract the weekday from a series of dates
        :param series:
        :param format: "%Y-%m-%d HH:mm:ss"
        :return:
        """
        return self.to_datetime(series, format=format).dt.weekday

    @staticmethod
    @abstractmethod
    def format_date(self, current_format=None, output_format=None):
        pass

    @staticmethod
    @abstractmethod
    def time_between(self, value=None, date_format=None):
        pass

    def years_between(self, series, value=None, date_format=None):
        return self.time_between(series, value, date_format).dt.days / 365.25

    def months_between(self, series, value=None, date_format=None):
        return self.time_between(series, value, date_format).dt.days / 30.436875

    def days_between(self, series, value=None, date_format=None):
        return self.time_between(series, value, date_format).dt.days

    def hours_between(self, series, value=None, date_format=None):
        series = self.time_between(series, value, date_format)
        return series.dt.days * 24.0 + series.dt.seconds / 3600.0

    def minutes_between(self, series, value=None, date_format=None):
        series = self.time_between(series, value, date_format)
        return series.dt.days * 1440.0 + series.dt.seconds / 60.0

    def seconds_between(self, series, value=None, date_format=None):
        series = self.time_between(series, value, date_format)
        return series.dt.days * 86400 + series.dt.seconds

    def domain(self, series):
        import hiurlparser
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["domain"], na_action=None)

    def top_domain(self, series):
        import hiurlparser
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["top_domain"], na_action=None)

    def sub_domain(self, series):
        import hiurlparser
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["sub_domain"], na_action=None)

    def url_scheme(self, series):
        import hiurlparser
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["protocol"], na_action=None)

    def url_path(self, series):
        import hiurlparser
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["path"], na_action=None)

    def url_file(self, series):
        import hiurlparser
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["file"], na_action=None)

    def url_query(self, series):
        import hiurlparser
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["query"], na_action=None)

    def url_fragment(self, series):
        import hiurlparser
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["fragment"], na_action=None)

    def host(self, series):
        import hiurlparser
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["host"], na_action=None)

    def port(self, series):
        import hiurlparser
        return self.to_string(series).map(lambda v: hiurlparser.parse_url(v)["port"], na_action=None)

    def email_username(self, series):
        return self.to_string_accessor(series).split('@').str[0]

    def email_domain(self, series):
        return self.to_string_accessor(series).split('@').str[1]

    def infer_data_types(self, value, cols_data_types):
        """
        Infer the data types.
        Please be aware that the order in which the value is checked is important and will change the final result
        :param value:
        :param cols_data_types:
        :return:
        """

        if is_list(value):
            dtype = ProfilerDataTypes.ARRAY.value
        elif is_null(value):
            dtype = ProfilerDataTypes.NULL.value
        elif is_bool(value):
            dtype = ProfilerDataTypes.BOOLEAN.value
        elif is_credit_card_number(value):
            dtype = ProfilerDataTypes.CREDIT_CARD_NUMBER.value
        elif is_zip_code(value):
            dtype = ProfilerDataTypes.ZIP_CODE.value
        elif is_int_like(value):
            dtype = ProfilerDataTypes.INT.value
        elif is_float_like(value):
            dtype = ProfilerDataTypes.FLOAT.value
        elif is_datetime(value):
            dtype = ProfilerDataTypes.DATETIME.value
        elif is_missing(value):
            dtype = ProfilerDataTypes.MISSING.value
        elif is_str(value):
            if is_datetime_str(value):
                dtype = ProfilerDataTypes.DATETIME.value
            elif is_ip(value):
                dtype = ProfilerDataTypes.IP.value
            elif is_url(value):
                dtype = ProfilerDataTypes.URL.value
            elif is_email(value):
                dtype = ProfilerDataTypes.EMAIL.value
            elif is_gender(value):
                dtype = ProfilerDataTypes.GENDER.value
            elif is_phone_number(value):
                dtype = ProfilerDataTypes.PHONE_NUMBER.value
            else:
                dtype = ProfilerDataTypes.STRING.value
        elif is_object_value(value):
            dtype = ProfilerDataTypes.OBJECT.value

        return dtype

    def date_formats(self, series):
        return series.map(lambda v: hidateinfer.infer([v]))

    def metaphone(self, series):
        return self.to_string(series).map(jellyfish.metaphone, na_action='ignore')

    def double_metaphone(self, series):
        return self.to_string(series).map(doublemetaphone, na_action='ignore')

    def nysiis(self, series):
        return self.to_string(series).map(jellyfish.nysiis, na_action='ignore')

    def match_rating_codex(self, series):
        return self.to_string(series).map(jellyfish.match_rating_codex, na_action='ignore')

    def soundex(self, series):
        return self.to_string(series).map(jellyfish.soundex, na_action='ignore')

    def levenshtein(self, series, other):
        if isinstance(other, str):
            return self.to_string(series).map(lambda v: jellyfish.levenshtein_distance(v, other))
        else:
            col_name = series.name
            other_name = other.name
            dfd = self.to_string(series).to_frame()
            dfd[other_name] = self.to_string(other)

            return dfd.apply(lambda d: jellyfish.levenshtein_distance(d[col_name], d[other_name]), axis=1).rename(
                col_name)
